#!/usr/bin/env python3

import argparse
import logging
import sys

from . import commands
from .util import exc_summary, setup_logging


_LOG = logging.getLogger(__name__)


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="cici",
        description="comic chat file toolkit",
        epilog="Made with love <3",
    )

    subparsers = parser.add_subparsers(
        title="Commands",
        dest="command",
        required=True,
        metavar="command",
        help="Description",
    )

    commands.info.register(subparsers)

    args = parser.parse_args()

    def _dispatch() -> None:
        setup_logging(args.debug)

        if args.command == "info":
            commands.info.dispatch(args)

    try:
        _dispatch()
    except KeyboardInterrupt:
        print("Interrupted via keyboard interrupt", file=sys.stderr)
        sys.exit(2)
    except (FileNotFoundError, PermissionError) as exc:
        print(str(exc), file=sys.stderr)
        sys.exit(3)
    except ExceptionGroup as exc_group:
        print(f"{str(exc_group)}:", file=sys.stderr)
        for exc_ in exc_group.exceptions:  # pylint: disable=not-an-iterable
            print(f"  - {exc_summary(exc_)}", file=sys.stderr)
            for note in exc_.__notes__:
                print(f"    ({note})", file=sys.stderr)
            # If we just re-raise the exception, we may not see the
            # full traces if there are too many, so log them instead
            # so we can be assured to see them in debug mode.
            _LOG.debug(exc_, exc_info=exc_)
        sys.exit(4)
    except Exception as exc:  # pylint: disable=broad-exception-caught
        if args.debug:
            raise
        print(f"Exiting with error ({exc_summary(exc)})", file=sys.stderr)
        sys.exit(5)
    except SystemExit as exc:
        if isinstance(exc.code, str):
            print(f"Exiting with error: {str(exc)}", file=sys.stderr)
            sys.exit(6)
        raise


if __name__ == "__main__":
    main()

# check [--debug] file[s]...
#  - implies strict/validate
#  - quiet by default? (no summary, emotions or images table?)
#  - maybe add -e/-i as an option?
#  - maaaaybe add -summary as an option?
#  - -Werror would be good here.

# repair <infile> <outfile>

# extract [--raw] [--all-poses] <infile> <outdir>



"""
This module lays out the "expected" image types.

This image metadata represents the image metadata of every avatar and
background file I have come across, split by context of where it
occurs. This is done so that I have an easy table to reference when I
want to know in what circumstances certain flags appear, or when
certain images are present or not and what forms they take, etc.

A certain type of image format being absent here doesn't mean it isn't
supported, but it may mean the tool will do something unpredictable
with it, because I haven't encountered it myself to test.

It's just kind of helpful to see this info all in one place.

Table entries are:

- bits per pixel
- Does the DIB header provide a byte size for the raw RGB data?
- Does the DIB header state the number of colors in the palette?
  (No defaults to 2^bpp)
- DIB header compression flag (0 is no compression.)
- How is the palette stored?
- Comic Chat v2.5 attributes field, if present.
"""

IMG_TABLE = {
    "icon": (
        # bpp, size?, ncol?, comp, PaletteType, flag
        (1, True, True, 0, "Embedded", None),
        (1, True, True, 0, "Implicit", 0x04),
        # ^ J Bruton, C Simitis, JL Dehaene, etc. MS Bug!
        (2, False, True, 0, "Implicit", 0x04),
        # Y-Guy, this is a bug! See `icon()`
        (4, True, True, 2, "Embedded", None),
        (4, False, True, 0, "Embedded", None),
        (4, True, True, 0, "Explicit", 0x02),
        (4, True, False, 0, "Explicit", 0x02),
        (4, True, False, 0, "Embedded", None),
        (4, True, False, 2, "Embedded", None),
        (8, False, True, 0, "Embedded", None),
        (8, True, True, 1, "Embedded", None),  # Oh! Neat. TOM. of course.
        (8, False, False, 0, "Embedded", None),
        (8, False, True, 0, "Explicit", 0x02),
        (8, True, True, 0, "Explicit", 0x02),
        (8, False, False, 0, "Explicit", 0x02),
        (24, False, False, 0, "None", None),
        (24, True, False, 0, "Empty", 0x02),
        (24, True, False, 0, "None", None),
        (24, False, False, 0, "Empty", 0x02),
    ),
    "bg": (
        # bpp, size?, ncol?, comp, PaletteType, flag
        (1, True, False, 0, "Explicit", 0x02),
        (4, True, True, 0, "Explicit", 0x02),
        (4, True, False, 0, "Explicit", 0x02),
        (4, False, True, 0, "Explicit", 0x02),
        (8, False, False, 0, "Explicit", 0x02),
        (8, False, True, 0, "Explicit", 0x02),
        (8, True, False, 0, "Explicit", 0x02),
        (8, True, True, 0, "Explicit", 0x02),
        (24, False, False, 0, "Empty", 0x02),
        (24, True, False, 0, "Empty", 0x02),
    ),
    "v55": [
        # First image: pose/head/face base image
        (
            # bpp, size?, ncol?, comp, PaletteType
            (1, False, True, 0, "Embedded"),
            (1, False, False, 0, "Embedded"),
            (1, True, True, 0, "Embedded"),
            (1, True, False, 0, "Embedded"),
            (4, False, True, 0, "Embedded"),
            (4, True, True, 2, "Embedded"),  # RLE4 compression
            (8, False, True, 0, "Embedded"),
            (8, True, True, 0, "Embedded"),
            (8, True, False, 0, "Embedded"),
            (24, False, False, 0, "None"),
            (24, True, False, 0, "None"),
        ),
        # Second image: compositing mask
        (
            # bpp, size?, ncol?, comp, PaletteType
            (1, True, False, 0, "Embedded"),
            (1, False, False, 0, "Embedded"),
            (1, False, True, 0, "Embedded"),
            (4, True, True, 2, "Embedded"),
        ),
        # Third image: halo mask
        (
            # bpp, size?, ncol?, comp, PaletteType
            (1, True, False, 0, "Embedded"),  # kwensa
            (1, True, False, 0, "Embedded"),  # Jean-Luc, John Bruton
            (1, True, True, 0, "Embedded"),  # Jean-Luc, John Bruton
            (1, False, True, 0, "Embedded"),  # 75 files
            (1, False, False, 0, "Embedded"),  # shizuka
            (4, True, True, 2, "Embedded"),  # tom
            (8, False, True, 0, "Embedded"),  # missvampire
            (24, True, False, 0, "None"),  # 7 files
            (24, False, False, 0, "None"),  # 69 files
        ),
    ],
    "v66": [
        # First image: ...
        (
            # bpp, size?, ncol?, comp, PaletteType, flag
            (2, False, True, 0, "Implicit", 0x04),
            (4, True, True, 0, "Explicit", 0x02),
            (4, False, True, 0, "Explicit", 0x02),
            (8, False, True, 0, "Explicit", 0x02),
            (8, True, True, 0, "Explicit", 0x02),
            (8, True, False, 0, "Explicit", 0x02),
            (24, False, False, 0, "Empty", 0x02),
            (24, True, False, 0, "Empty", 0x02),
        ),
        # Second image: ...
        (
            # bpp, size?, ncol?, comp, PaletteType, flag
            (2, False, True, 0, "Implicit", 0x05),
        ),
        # Third image: ...
        (
            # bpp, size?, ncol?, comp, PaletteType, flag
            (1, False, True, 0, "Implicit", 0x03),
        ),
    ],
}



"""
Binary - Microsoft Comic Chat Binary files support

This module provides access to the `Binary` class, which provides
support for reading and writing Microsoft Comic Chat binary files,
including AVB (Avatar) and BGB (Binary) files.
"""

import enum
import hashlib
from io import BytesIO
import logging
import os
from pathlib import Path
import struct
from typing import (
    BinaryIO,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from PIL import Image
import chardet

from .bitmap import Bitmap, CompressedBitmap
from .emotion import EmotionHeaderBase
from .imgmeta import ImageMetadata
from .table import IMG_TABLE
from .util import exc_summary, multi_exception_shim


_LOG = logging.getLogger(__name__)


class FileVersion(enum.Enum):
    CCHAT_21 = 1
    CCHAT_25 = 2

    def version(self) -> float:
        if self == FileVersion.CCHAT_21:
            ret = 2.1
        else:
            assert self == FileVersion.CCHAT_25
            ret = 2.5
        return ret


class FileType(enum.Enum):
    # Combined, static poses. This is the format produced by the
    # widely available CCEdit v55 and v66 tools.
    AVATAR_STATIC = 1

    # Dynamic poses. This is the format used by most
    # Microsoft-published avatars, but a few enterprising hackers
    # managed to create their own; notably Shunji Haruki.
    AVATAR_DYNAMIC = 2

    # Background files: Only available in CCHAT_25 files.
    BACKGROUND = 3

    def describe(self) -> str:
        if self == FileType.AVATAR_STATIC:
            ret = "Avatar, Static poses"
        elif self == FileType.AVATAR_DYNAMIC:
            ret = "Avatar, Dynamic poses"
        else:
            assert self == FileType.BACKGROUND
            ret = "Background"
        return ret


class Field(enum.Enum):
    NAME = 1
    COMPOSITION = 2
    ICON_OFFSET = 3
    FACES_V1 = 4
    BODIES_V1 = 5
    END_OF_HEADER = 6
    END_OF_FILE = 7
    EIGHT = 8  # What's this do? I dunno :3
    COMBINED_V1 = 9
    FACES_V2 = 0x0A
    BODIES_V2 = 0x0B
    COMBINED_V2 = 0x0C
    ICON_INFO = 0x100
    BG_INFO = 0x102
    COPYRIGHT = 0x103
    DOWNLOAD = 0x104
    DOWNLOAD2 = 0x105
    PROTECTION = 0x106
    REBASE = 0x107


def decodify(raw: bytes) -> str:
    """
    decodify tries to decode string data.

    AVB/BGB files do not store what encoding they use, so it's a trial
    and error process. The large majority of files accessible to me,
    an English speaker, are likely ASCII or Windows-1252/ISO-8859-1.

    In some cases, this is ambiguous and some files may actually be
    encoded with other encodings; we'll try our best to guess a
    reasonable default.
    """
    # Try as UTF-8 first. This will probably only work when it's
    # literal ASCII.
    try:
        return raw.decode()
    except UnicodeDecodeError:
        pass

    # It is probably Windows-1252/ISO-8859-1 if the only non-ASCII
    # characters are © and/or ®. These characters could be inserted
    # into the text field of the CCEdit editor with dedicated
    # clickable buttons, so their presence is likely.
    raw2 = raw.replace(b"\xAE", b"").replace(b"\xA9", b"")
    if raw2.isascii():
        return raw.decode("Windows-1252")

    # chardet never seems to guess Japanese, even when it is. For
    # now, "just try it". Comic Chat was very popular in Japan.
    try:
        return raw.decode("shift_jis")
    except UnicodeDecodeError:
        # Well, it wasn't Japanese.
        pass

    # Get a guess from chardet as a last resort.
    # Sometimes it throws its hands up and says "I dunno",
    # in which case, just default to Windows-1252 again.
    guess = chardet.detect(raw).get("encoding", "Windows-1252")
    if guess is None:
        guess = "Windows-1252"
    try:
        return raw.decode(guess)
    except UnicodeDecodeError:
        # ah geez. maybe it guessed something that isn't Windows-1252
        # and its own guess failed.
        pass

    return raw.decode("Windows-1252")


def encodify(value: str) -> bytes:
    """
    encodify is the opposite of `decodify` and outputs encoded bytes.

    Like decodify, this will try to guess the most appropriate
    encoding for a string.
    """
    # The nice thing about *encoding* is that it's a lot easier to
    # determine the encoding. If an encoding isn't legal, we try a
    # different one!

    # Try as ASCII first.
    try:
        return value.encode("ASCII") + b"\x00"
    except UnicodeEncodeError:
        pass

    # Then Shift_JIS (Japanese):
    try:
        return value.encode("Shift_JIS") + b"\x00"
    except UnicodeEncodeError:
        pass

    # Finally, Windows-1252.
    return value.encode("Windows-1252") + b"\x00"


class Binary:
    """
    This represents CChat AVB and BGB files.

    :param file: Can be None, bytes, a file-like object,
                 or a str/Path pointing to a file.

    :param parse_images: Set to False if you don't want to attempt
        parsing of images at all. This can be useful for parsing
        metadata out of damaged files. Images can be buffered post-hoc
        by calling `read_images()`, but the file will still need to be
        open. This is not supported when passing a str/path as @file.
    """

    def __init__(
        self,
        file: Union[bytes, str, Path, BinaryIO],
        parse_images: bool = True,
        autorepair: bool = False,
    ) -> None:
        # Settings
        self._autorepair = autorepair

        # Main Header
        self.version: Tuple[int, int]
        self.ftype: FileType
        self.fvers: FileVersion

        # Simple metadata
        self.name: Optional[str] = None
        self.copyright: Optional[str] = None
        self.author: Optional[str] = None
        self.download_orig: Optional[str] = None
        self.download_new: Optional[str] = None
        self.protected = False
        self._unk8: Optional[int] = None  # always 0x01 for avatars; absent otherwise
        self.composition: Optional[int] = None  # 0x05: body/pose-first, 0x02: head-first, 0x00: ??
        self.added_bytes = 0

        # Images and Emotion data:
        self._icon = ImageMetadata(self, "ICON", "Icon")
        self._bg = ImageMetadata(self, "BKGD", "Background")
        self.emotions: List[EmotionHeaderBase] = []

        # Fields for strictness checking, debugging, and summary output:
        self._regions: List[Tuple[int, int]] = []
        self._fields_seen: List[int] = []
        self._end_of_header = 0
        self._end_of_bitmaps = 0
        self._img_table: Optional[List[str]] = None
        self._canvas_size: Optional[Tuple[int, int]] = None
        self._eof_checked = False

        # Alright, let's get this show on the road.

        # NB: It'd be nice to allow empty Binary() files and to fill them out as we go,
        # but the class isn't *quite* set up for that just yet, we need some refactoring.
        # For now, (sorry), require something to be passed in.
        self._file: BinaryIO
        if isinstance(file, bytes):
            self._file = BytesIO(file)
            self.read(parse_images)
        elif isinstance(file, (str, Path)):
            if not parse_images:
                raise ValueError(
                    "parse_images=False not supported when file is passed as str/Path."
                )
            with open(file, "rb") as infile:
                self._file = infile
                self.read(parse_images)
        else:
            self._file = file
            self.read(parse_images)

    def _warn(self, msg: str) -> None:
        _LOG.warning(msg)

    def _read_string(self, length_prefix: bool = False, what: str = "") -> str:
        """
        Read string data from the open AVB/BGB file at current offset.

        Strings associated with names don't store a length prefix, but
        every other string does. This method maintains bug-level
        compatibility with how CChat reads strings, which ignores the
        length prefix stored in the file.

        This method calls `decodify` and returns a decoded string
        without a trailing null byte.
        """
        n = 0
        if length_prefix:
            n = struct.unpack("<h", self._file.read(2))[0]

        # Quirk: even if we have a length prefix, cchat reads until
        #        null regardless of the length. SCARY!
        #        This can lead to both under- and overreads.
        data = bytearray()
        while True:
            data += self._file.read(1)
            if data[-1] == 0:
                break

        if length_prefix and n != len(data):
            self._warn(f"String read for {what} got {len(data)} bytes, was expecting {n}!")

        return decodify(data[:-1])

    def _read_length_prefixed_data(self) -> bytes:
        """
        Read a length-prefixed data segment at the current offset.

        This cannot be used for length-prefixed strings due to a bug
        in MSCC, see `_read_string`.
        """
        n = struct.unpack("<h", self._file.read(2))[0]
        return self._file.read(n)

    def _read_copyright_author(self) -> None:
        """
        Read the copyright and author from file at the current offset.
        """
        copyauth = self._read_string(length_prefix=True, what="Copyright/Author")
        if "\\n" in copyauth:
            self.copyright, self.author = copyauth.split("\\n", maxsplit=1)
        else:
            self.copyright = copyauth

    def _seek(self, offset: int, whence: int) -> int:
        curpos = self._file.tell()
        pos = self._file.seek(offset, whence)
        if pos > self._size:
            self._file.seek(curpos, os.SEEK_SET)
            raise EOFError(f"Offset 0x{pos:02X} is beyond EOF (0x{self._size:02X})")
        return pos

    def read(self, parse_images: bool = True) -> None:
        """
        Parse the open AVB/BGB file from the current offset.

        If parse_images is False, images will not be parsed or
        buffered at all.  Call `read_images` later if you wish to do
        it separately, but keep in mind that the file will still need
        to be open!

        This option is most useful when dealing with damaged files
        that need to be repaired, or for getting metadata out of
        damaged files.

        Also note: when parse_images is False, warnings related to EOF
        markers and extra trailing data will not be triggered unless
        `read_images` is later called.
        """
        curpos = self._file.tell()
        self._size = self._file.seek(0, os.SEEK_END)
        self._file.seek(curpos, os.SEEK_SET)

        magic = self._file.read(2)
        _LOG.info("signature/magic: 0x%s", magic.hex())

        if magic == b"\x81\x00":
            self.version = (2, 1)
        elif magic == b"\x81\x81":
            self.version = (2, 5)
        else:
            raise ValueError(
                "Unrecognized file magic; expected 0x8181 (v2.5) or 0x8100 (v2.1); got "
                f"0x{magic.hex()}"
            )

        self.ftype = FileType(struct.unpack("<h", self._file.read(2))[0])
        _LOG.info("type: %s (0x%04x)", self.ftype, self.ftype.value)
        self.fvers = FileVersion(struct.unpack("<h", self._file.read(2))[0])
        _LOG.info("version: %s (0x%04x)", self.fvers, self.fvers.value)

        fingerprint = (magic, self.ftype, self.fvers)
        if fingerprint not in (
            (b"\x81\x81", FileType.AVATAR_STATIC, FileVersion.CCHAT_25),
            (b"\x81\x81", FileType.AVATAR_DYNAMIC, FileVersion.CCHAT_25),
            (b"\x81\x81", FileType.BACKGROUND, FileVersion.CCHAT_25),
            (b"\x81\x00", FileType.AVATAR_STATIC, FileVersion.CCHAT_21),
            (b"\x81\x00", FileType.AVATAR_DYNAMIC, FileVersion.CCHAT_21),
        ):
            raise ValueError(f"Unexpected file signature: {fingerprint}")

        fields_seen = []
        while True:
            field_type = struct.unpack("<h", self._file.read(2))[0]
            fields_seen.append(field_type)
            try:
                field = Field(field_type)
            except:
                _LOG.error(
                    "0x%x is not a valid metadata field type (@offset 0x%x)",
                    field_type,
                    self._file.tell(),
                )
                raise

            if field == Field.NAME:
                self.name = self._read_string()
                _LOG.info("0x0001 / name: %s", self.name)

            elif field == Field.COMPOSITION:
                self.composition = struct.unpack("<h", self._file.read(2))[0]
                _LOG.info("0x0002 / composition flag: 0x%02x", self.composition)

            elif field == Field.ICON_OFFSET:
                # This is used for v2.1 files only. ICON_INFO is used otherwise.
                offset = struct.unpack("<I", self._file.read(4))[0]
                self._icon.offset = offset
                _LOG.info("0x0003 / icon offset: 0x%08x", self._icon.offset)
                if not self._icon.offset:
                    self._warn("Icon offset was zero. That's probably not right.")

            elif field == Field.EIGHT:
                self._unk8 = struct.unpack("<h", self._file.read(2))[0]
                _LOG.info("0x0008 / unknown avatar field: 0x%02x", self._unk8)

            # All 0x1nn field types are length-prefixed.
            elif field == Field.ICON_INFO:
                data = self._read_length_prefixed_data()
                if len(data) != 6:
                    _LOG.debug("Bad header data: %s", repr(data))
                    raise ValueError("Icon header size was not exactly 6 bytes")

                (self._icon.offset, self._icon.unk, self._icon.flag) = struct.unpack("<IBB", data)

                _LOG.info("0x0100 / icon header:")
                _LOG.info("  offset:     0x%08x", self._icon.offset)
                _LOG.info("  unknown:    0x%02x", self._icon.unk)
                _LOG.info("  image type: 0x%02x", self._icon.flag)
                if not self._icon.offset:
                    self._warn("Icon offset was zero. That's probably not right.")

            elif field == Field.BG_INFO:
                data = self._read_length_prefixed_data()
                if len(data) != 6:
                    _LOG.debug("Bad header data: %s", repr(data))
                    raise ValueError("Background header size was not exactly 6 bytes")

                (self._bg.offset, self._bg.unk, self._bg.flag) = struct.unpack("<IBB", data)

                _LOG.info("0x0102 / background header:")
                _LOG.info("  offset:     0x%08x", self._bg.offset)
                _LOG.info("  unknown:    0x%02x", self._bg.unk)
                _LOG.info("  image type: 0x%02x", self._bg.flag)
                if not self._bg.offset:
                    self._warn("Background offset was zero. That's probably not right.")

            elif field == Field.COPYRIGHT:
                self._read_copyright_author()
                _LOG.info("0x0103 / authorship information:")
                _LOG.info("  copyright: %s", self.copyright)
                _LOG.info("  author: %s", self.author)

            elif field == Field.DOWNLOAD:
                self.download_orig = self._read_string(length_prefix=True, what="Download URL")
                _LOG.info("0x0104 / download URL: %s", self.download_orig)

            elif field == Field.DOWNLOAD2:
                self.download_new = self._read_string(length_prefix=True, what="Download URL2")
                _LOG.info("0x0105 / download override: %s", self.download_new)

            elif field == Field.PROTECTION:
                data = self._read_length_prefixed_data()
                if len(data) != 1:
                    _LOG.debug("Bad protection header data: %s", repr(data))
                    raise ValueError("Protection header size was not exactly one byte")

                self.protected = struct.unpack("<?", data)[0]
                _LOG.info("0x0106 / protected: %s", self.protected)

            elif field == Field.REBASE:
                data = self._read_length_prefixed_data()
                if len(data) != 4:
                    _LOG.debug("Bad rebase header data: %s", repr(data))
                    raise ValueError("Rebase header was not exactly 4 bytes")

                offset = struct.unpack("<i", data)[0]
                _LOG.info("0x0107 / rebase: %d byte(s)", offset)
                self.added_bytes += offset

            elif field in (
                Field.FACES_V1,
                Field.BODIES_V1,
                Field.COMBINED_V1,
                Field.FACES_V2,
                Field.BODIES_V2,
                Field.COMBINED_V2,
            ):
                count = struct.unpack("<h", self._file.read(2))[0]
                _LOG.info(
                    "0x%04x / %s: %d entries",
                    field_type,
                    EmotionHeaderBase.describe_field(field_type),
                    count,
                )
                klass = EmotionHeaderBase.get_klass(field_type)
                combined = field in (Field.COMBINED_V1, Field.COMBINED_V2)

                hdr = klass.table_header(combined)
                _LOG.info("  %s", hdr)
                _LOG.info("  %s", "-" * len(hdr))

                for i in range(count):
                    eheader = klass(self, field_type, i)
                    eheader.read(self._file)
                    self.emotions.append(eheader)
                    _LOG.info("  %s", eheader.get_summary_row())

            elif field == Field.END_OF_HEADER:
                _LOG.info("0x0006 / End of header; start of image data")
                break

            else:
                raise ValueError(f"Unknown field type {field_type:X}")

        _LOG.info("header fingerprint: %s", self.header_fingerprint())

        self._end_of_header = self._file.tell()
        _LOG.info(f"end of header: 0x{self._end_of_header:X}")
        self._regions.append((0, self._end_of_header))
        self._fields_seen = fields_seen
        if parse_images:
            self.read_images()

    def read_images(self, evaluate: bool = False) -> None:
        self._parse_images(evaluate)
        # We don't know where the image table ends unless we can parse it,
        # so this check winds up here as the only place it *can* go.
        if not self._eof_checked:
            self._check_end_of_file()

    def print_summary(self, advanced: bool = False) -> None:
        output = {}

        if hasattr(self._file, "name"):
            output["file"] = self._file.name
            output["size"] = str(os.path.getsize(self._file.name))
        elif hasattr(self._file, "getvalue"):
            output["file"] = "<IO Buffer>"
            output["size"] = str(len(self._file.getvalue()))
        else:
            output["file"] = "<No File>"
            output["size"] = "<Unknown>"

        if advanced:
            output["type"] = repr(self.ftype)
            output["version"] = repr(self.fvers)
        else:
            output["type"] = self.ftype.describe()
            output["version"] = str(self.fvers.version())

        # Hack to get a little separator O:-)
        output[" "] = " "

        output["name"] = self.name or "--"
        output["copyright"] = self.copyright or "--"
        output["author"] = self.author or "--"

        output["download (new)"] = self.download_new or "--"
        output["download (old)"] = self.download_orig or "--"
        output["protected"] = "yes" if self.protected else "no"

        # O:-)
        output["  "] = " "

        output["emotions"] = f"{len(self.emotions)}"
        output["unique images"] = str(
            len(set(emotion.unstable_hash() for emotion in self.emotions))
        )
        output["emotion types"] = str(len(set(e.emotion for e in self.emotions))) + "/12"  # lol
        output["fingerprint"] = self.header_fingerprint()

        width = max(len(k) for k in output) + 1
        for k, v in output.items():
            if k.strip():
                k = k + ":"
            print(f"{k:{width}} {v}")

    def print_emotion_table(self) -> None:
        combined = not any(e.face_data() or e.body_data() for e in self.emotions)
        last_record_type = None
        for emotion in self.emotions:
            if emotion.field_type != last_record_type:
                last_record_type = emotion.field_type
                print("")
                hdr = emotion.table_header(combined)
                print(hdr)
                print("-" * len(hdr))
            print(emotion.get_summary_row())

    def header_fingerprint(self) -> str:
        ret: List[object] = [
            self.ftype.value,
            self.fvers.value,
            self.composition,
            self._icon.unk,
            self._icon.flag,
            self._bg.unk,
            self._bg.flag,
        ]
        for emotion in self.emotions:
            ret.append(emotion.fingerprint())
        data = str(tuple(ret)).encode()
        return hashlib.md5(data).hexdigest()

    def _parse_images(self, evaluate: bool = False) -> None:
        """
        Parse all of the image metadata present in this file.

        This method is called towards the end of read(). This method
        will parse all of the image metadata, but does not actually
        process or reconstruct any of those images yet. Data will be
        *buffered* but not decompressed.

        This method may report failures if the offsets listed in the
        headers are incorrect, the file is short, the bitmap headers
        in the file are corrupt, etc. When evaluate is False, this
        method may surprisingly sometimes *not fail* for CChat 2.5
        files with zlib stream corruption, or other image data
        interpretation problems which may only surface when fully
        evaluating images.

        On Python3.11+, ExceptionGroup will be raised to report *all*
        image parsing problems instead of stopping at only the first
        problem.
        """
        exceptions = []
        img_table = []

        def _handle_row(entry: Tuple[str, Optional[Exception]]) -> None:
            row, exc = entry
            if row:
                _LOG.info(row)
                img_table.append(row)
            if exc:
                assert row
                exceptions.append(exc)

        # ---

        _LOG.info("")
        _LOG.info("Image data:")
        for line in ImageMetadata.img_table_hdr():
            _LOG.info(line)
            img_table.append(line)

        _handle_row(self._icon.img_table_row(evaluate))
        _handle_row(self._bg.img_table_row(evaluate))

        for emotion in self.emotions:
            problems = False
            for entry in emotion.get_img_table_rows(evaluate):
                _handle_row(entry)
                problems = problems or bool(entry[1])
            if evaluate and not problems:
                try:
                    # Final sanity check!
                    _ = emotion.render()
                except Exception as exc:  # pylint: disable=broad-exception-caught
                    _handle_row(
                        (f"    ❌ Problem compositing {emotion.kind()}: {exc_summary(exc)}", exc)
                    )

        self._img_table = img_table
        multi_exception_shim(exceptions, _LOG, "There were multiple problems parsing images")

    def print_image_table(self) -> None:
        if self._img_table is None:
            raise RuntimeError("read_images() must be called before printing the image table.")
        for line in self._img_table:
            print(line)

    def _check_end_of_file(self) -> None:
        """
        This method checks that there is no extra data at the end of the file.

        It's called at the end of `read_images()`. It will only ever
        produce warnings, it doesn't raise any exceptions.
        """
        start = self._end_of_bitmaps
        self._file.seek(start, os.SEEK_SET)

        data = self._file.read(2)
        if data != b"\x07\x00":
            self._warn(f"0x07 EOF record not found immediately after image data (at 0x{start:X}).")
            self._file.seek(start, os.SEEK_SET)

        junk_start = self._file.tell()
        data = self._file.read()
        if data == b"":
            # EOF! We love EOF.
            pass
        elif data.replace(b"\x00", b""):
            self._warn(f"Found junk data at end of file (from 0x{junk_start:X}).")
        else:
            self._warn(f"Found empty bytes at end of file (from 0x{junk_start:X}).")

        end = self._file.tell()
        self._regions.append((start, end))
        self._eof_checked = True

    def get_offset(self, offset: int) -> int:
        """
        Get an absolute offset from a relative one.

        Offsets advertised in AVB/BGB files can be rebased if 0x107
        headers are present. This method translates the recorded
        offset into the real, absolute offset.
        """
        return offset + self.added_bytes

    def get_bmp(self, offset: int, flags: int = 0) -> Bitmap:
        """
        Get a Bitmap object located at a given offset.

        This method takes a raw offset as advertised directly in the
        Emotion/Pose data tables and it may not match the real offset
        in the binary file if any rebase headers are present in the
        file header.

        The object returned may be a Bitmap or a CompressedBitmap.

        Raw data will be buffered but not deserialized into a usable
        format; this method may error out if the data stream is
        truncated or if the bitmap headers are corrupted in some way.
        """
        ofs = self.get_offset(offset)
        self._seek(ofs, os.SEEK_SET)

        bmp: Bitmap
        if flags:
            bmp = CompressedBitmap(self._file, autorepair=self._autorepair)
            bmp.generate_implied_palette(flags)
        else:
            bmp = Bitmap(self._file, autorepair=self._autorepair)

        end = self._file.tell()
        if (ofs, end) not in self._regions:
            self._regions.append((ofs, end))
        self._end_of_bitmaps = max(self._end_of_bitmaps, end)

        return bmp

    @property
    def background(self) -> Optional[Image.Image]:
        # A setter API would be nice to add (...)
        return self._bg.img

    @property
    def icon(self) -> Optional[Image.Image]:
        # A setter API would be nice to add (...)
        return self._icon.img

    def strictness_check(self) -> None:
        # A lot of these checks are redundant as the library has grown
        # and become more robust; I simply haven't audited them all to
        # see which are now superfluous. In the worst case, they serve
        # as a convenient localized reference for the assumptions this
        # library makes.
        #
        # Ideally, these would all be warnings; but at time of writing
        # I have only 7 files in my possession that trigger any
        # assertions (without running into a hard error earlier in the
        # parse), and it's still useful to keep these assertions as
        # strict failures for purposes of understanding which
        # circumstances *never* happen and which are indicative of
        # more minor problems. (IOW: as warnings, they get lost in the
        # noise of more minor problems. For remaining assertion
        # failures witnessed on files in circulations, they should
        # either be converted to hard errors if they are not usable in
        # comic chat, or downgraded to warnings if they DO work, but
        # potentially with issues.)

        # ###############################################
        # *** Validate metadata fields present/absent ***
        # ###############################################

        # These fields must be present:
        mandatory = {
            FileType.AVATAR_DYNAMIC: {
                FileVersion.CCHAT_21: {0x04, 0x05},
                FileVersion.CCHAT_25: {
                    0x0A,
                    0x0B,
                    0x01,
                    0x02,
                    0x100,
                },  # 0x08: downgraded to warning; observed on files in circulation
            },
            FileType.AVATAR_STATIC: {
                FileVersion.CCHAT_21: {0x09},
                FileVersion.CCHAT_25: {0x0C, 0x01, 0x02, 0x100},  # 0x08: warn-only
            },
            FileType.BACKGROUND: {FileVersion.CCHAT_25: {0x102}},
        }

        # These fields must be absent:
        forbidden = set()
        if self.ftype == FileType.AVATAR_DYNAMIC:
            forbidden |= {0x09, 0x0C}
        if self.ftype == FileType.AVATAR_STATIC:
            forbidden |= {0x04, 0x05, 0x0A, 0x0B}
        if self.ftype == FileType.BACKGROUND:
            forbidden |= {
                0x01,
                0x02,
                0x03,
                0x04,
                0x05,
                0x08,
                0x09,
                0x0A,
                0x0B,
                0x0C,
                0x100,
            }
        if self.fvers == FileVersion.CCHAT_21:
            forbidden |= {0x0A, 0x0B, 0x0C}
            forbidden |= {0x100, 0x103, 0x104, 0x105, 0x106, 0x107}
        if self.fvers == FileVersion.CCHAT_25:
            forbidden |= {0x04, 0x05, 0x09}
            forbidden |= {0x03}

        for field in forbidden:
            assert field not in self._fields_seen, (
                f"field 0x{field:02X} unexpectedly present in file type "
                f"{self.ftype.name}/{self.fvers.name}"
            )
        for field in mandatory[self.ftype][self.fvers]:
            assert field in self._fields_seen, (
                f"field 0x{field:02X} unexpectedly missing in file type "
                f"{self.ftype.name}/{self.fvers.name}"
            )

        if self.fvers == FileVersion.CCHAT_25 and self.ftype != FileType.BACKGROUND:
            if 0x08 not in self._fields_seen:
                self._warn(
                    "Field 0x08 missing from binary; This is likely caused by hex-editing damage."
                )

        # ##############################
        # *** Validate Misc Fields *** #
        # ##############################

        if self.fvers == FileVersion.CCHAT_25 and self.ftype != FileType.BACKGROUND:
            assert self._icon.unk == 0x01, f"Icon field #3 was not 0x01 (0x{self._icon.unk:02X})"
            assert self._icon.flag in (0x02, 0x04), self._icon.flag

        if self.ftype == FileType.AVATAR_DYNAMIC:
            # Only SUSAN.AVB (2.1 or 2.5) uses 0x02 here.
            # Absolutely everything else uses 0x05.
            assert self.composition in (
                0x02,
                0x05,
            ), f"Expecting composition type 0x05 or 0x02, found 0x{self.composition:02X}"
        elif self.ftype == FileType.AVATAR_STATIC and self.fvers == FileVersion.CCHAT_21:
            # glenda.avb and tux.avb have a 0x00 here. Don't know why.
            assert self.composition in (
                0x00,
                0x05,
            ), f"Expecting composition type 0x05 or 0x00, found 0x{self.composition:02X}"
        elif self.ftype != FileType.BACKGROUND:
            assert (
                self.composition == 0x05
            ), f"Expecting composition type 0x05, found 0x{self.composition:02X}"

        if self.ftype == FileType.BACKGROUND:
            assert self._bg.unk == 0x01
            assert self._bg.flag == 0x02
        else:
            assert self._icon.offset
            assert (
                self._unk8 == 0x01 or 0x08 not in self._fields_seen
            ), f"Expecting 0x01 in field 0x08, got 0x{self._unk8:X}"

        # ####################################
        # *** Validate Images & Emotions *** #
        # ####################################

        for emotion in self.emotions:
            try:
                emotion.strictness_check()
            except:
                _LOG.error("strictness check failed for %s", emotion.describe())
                raise

        if bmp := self._icon.bmp:
            info = bmp.characterize((self._icon.flag,))
            assert info in IMG_TABLE["icon"], f"unexpected image info for icon: {info}"

        if bmp := self._bg.bmp:
            info = bmp.characterize((self._bg.flag,))
            assert info in IMG_TABLE["bg"], f"unexpected image info for bgb: {info}"

        # ##############################
        # *** Validate File Layout *** #
        # ##############################

        if self.ftype == FileType.BACKGROUND:
            assert self._bg.offset is not None
            expected_offset = self.get_offset(self._bg.offset)
        else:
            assert self._icon.offset is not None
            expected_offset = self.get_offset(self._icon.offset)
        assert self._end_of_header == expected_offset, (
            f"End of header (0x{self._end_of_header:02x})"
            f" is not located at icon/bg offset (0x{expected_offset:02x})"
        )

        self._regions = sorted(self._regions)
        end = 0
        for region in self._regions:
            if region[0] == end:
                end = region[1]
        if end != self._file.tell():
            for region in self._regions:
                _LOG.debug("%x %x", region[0], region[1])
            self._warn(f"Detected unused bytes starting at 0x{end:X}")

    def guess_canvas_size(self) -> Tuple[int, int]:
        if self._canvas_size is not None:
            return self._canvas_size
        max_height = 0
        max_width = 0
        for emotion in self.emotions:
            # Every last emotion header; pose/body/face should have a base image.
            # buut, sometimes files are corrupt and we can't read it.
            try:
                bmp = emotion.bitmap(0)
                if not bmp:
                    continue
            except Exception as exc:  # pylint: disable=broad-exception-caught
                _LOG.warning(f"Error reading image while guessing canvas size: {exc_summary(exc)}")
                _LOG.debug("Error reading image while guessing canvas size", exc_info=exc)
                continue

            max_width = max(max_width, bmp.dib.width)
            max_height = max(max_height, bmp.dib.height)
        self._canvas_size = (max_width, max_height)
        return self._canvas_size

    @classmethod
    def _write_field(cls, file: BinaryIO, field: Field, value: bytes = b"") -> None:
        file.write(struct.pack("<h", field.value))
        if field.value >= 0x100:
            # All fields in this range are length-prefixed.
            file.write(struct.pack("<h", len(value)))
        file.write(value)

    def _write_image_fields(self, file: BinaryIO) -> None:
        """
        Write out the image fields to a file.

        This includes BG, Icon, emotions, and the end-of-header record.
        """
        if self._bg.offset:
            # Background Info (V2)
            self._write_field(
                file,
                Field.BG_INFO,
                struct.pack("<IBB", self._bg.offset, self._bg.unk, self._bg.flag),
            )

        if self._icon.offset:
            if self.fvers == FileVersion.CCHAT_21:
                # Icon Offset (V1)
                self._write_field(file, Field.ICON_OFFSET, struct.pack("<I", self._icon.offset))
            else:
                # Icon Info (V2)
                self._write_field(
                    file,
                    Field.ICON_INFO,
                    struct.pack("<IBB", self._icon.offset, self._icon.unk, self._icon.flag),
                )

        # Emotion headers
        for field_type in (
            Field.FACES_V1,
            Field.BODIES_V1,
            Field.COMBINED_V1,
            Field.FACES_V2,
            Field.BODIES_V2,
            Field.COMBINED_V2,
        ):
            emotions = [e for e in self.emotions if e.field_type == field_type.value]
            if not emotions:
                continue

            self._write_field(file, field_type, struct.pack("<h", len(emotions)))
            for emotion in emotions:
                file.write(bytes(emotion))

        # END_OF_HEADER
        self._write_field(file, Field.END_OF_HEADER)

    def write(self, file: BinaryIO) -> None:
        """
        Write this binary back out to file.

        Offsets are recomputed and re-written to file; offsets of
        emotion headers **will be altered in-memory** to reflect the
        new file layout. Rebase headers are removed.

        Any data that was repaired in-memory is written back out to
        file in place of the original corrupted data, in effect
        repairing the file.

        Any unused data regions between image files, or any unused
        trailing data will also be omitted.
        """

        # Gotta make sure the image table is fully cooked before we
        # attempt to write. Evaluation is not necessary; we'll just
        # write garbage data back out if we got garbage in.
        if self._img_table is None:
            self.read_images()

        if self.version == (2, 1):
            magic = b"\x81\x00"
        else:
            magic = b"\x81\x81"
        header = struct.pack("<2shh", magic, self.ftype.value, self.fvers.value)
        file.write(header)

        # Field.REBASE would get written here if we were going to do it.
        # No need, we're re-generating the file from scratch with freshly
        # computed correct offsets.
        self.added_bytes = 0

        if self.ftype in (FileType.AVATAR_STATIC, FileType.AVATAR_DYNAMIC):
            # Name is not included for FileType.BACKGROUND
            name = self.name or "NoName"
            data = encodify(name)
            self._write_field(file, Field.NAME, data)
            self._write_field(file, Field.EIGHT, struct.pack("<h", self._unk8 or 0x01))
            self._write_field(file, Field.COMPOSITION, struct.pack("<h", self.composition))

        if self.fvers == FileVersion.CCHAT_25:
            if self.copyright or self.author:
                copyright_ = self.copyright or "Not specified"
                author = self.author or "Not specified"
                copyauth = f"{copyright_}\\n{author}"
                data = encodify(copyauth)
                self._write_field(file, Field.COPYRIGHT, data)
            if self.download_orig:
                data = encodify(self.download_orig)
                self._write_field(file, Field.DOWNLOAD, data)
            if self.protected:
                self._write_field(file, Field.PROTECTION, struct.pack("<B", self.protected))
            if self.download_new:
                data = encodify(self.download_new)
                self._write_field(file, Field.DOWNLOAD2, data)

        # Write out a mock-version of the rest of the header data with
        # (possibly) wrong offsets so we can compute where the data
        # region will start.

        dummy_buffer = BytesIO()
        self._write_image_fields(dummy_buffer)
        start_of_data = file.tell() + dummy_buffer.tell()
        del dummy_buffer

        # Now, plan where all the images will go...
        image_table: Dict[Tuple[int, str], int] = {}
        image_buffer = bytearray()

        def _write_image(bmp: Optional[Bitmap]) -> Optional[int]:
            if not bmp:
                return None

            # Deduplicate identical images O:-)
            data = bytes(bmp)
            ident = (len(data), hashlib.md5(data).hexdigest())

            if ident in image_table:
                return image_table[ident]

            offset = start_of_data + len(image_buffer)
            image_table[ident] = offset
            image_buffer.extend(data)
            return offset

        # Write out images to buffer and update header offsets
        if bmp := self._bg.bmp:
            self._bg.offset = _write_image(bmp)

        if bmp := self._icon.bmp:
            self._icon.offset = _write_image(bmp)

        for emotion in self.emotions:
            for i in range(3):
                emotion.update_offset(i, _write_image(emotion.bitmap(i)) or 0)

        # Now that the offsets have been updated, actually write out the rest of the header.
        self._write_image_fields(file)
        assert file.tell() == start_of_data, "Miscalculated the header size - this is a bug :("

        # IMAGE DATA --
        file.write(image_buffer)

        # END_OF_FILE
        self._write_field(file, Field.END_OF_FILE)



#!/usr/bin/env python3

import argparse
import logging
from typing import Any

from cici.binary import Binary
from cici.util import capture_warnings

from ._common import add_common_arguments, open_file


_LOG = logging.getLogger(__name__)


_STRICT_MSG = """
This may be due to a poorly authored or edited file, or it might be due to corruption.
In some cases, it might be that this tool needs to be updated to support this file.
Please send this file to nago for analysis!
"""

# info [--strict] [--debug] [--e/emotions] [--i/images] [-v/--verbose] file[s]...

# --show-warnings: on by default
# add a -Werror flag to treat warnings as errors for sake of return code; implies show-warnings
# make --strict imply -Werror


def info(
    filename: str,
    strict: bool,
    show_warnings: bool,
    show_emotions: bool = False,
    show_images: bool = False,
) -> None:

    def _print_summary_table(avb: Binary) -> None:
        print("")
        print("Summary")
        print("=======")
        print("")
        avb.print_summary(advanced=False)
        print("")

    def _print_emotion_table(avb: Binary) -> None:
        if show_emotions:
            print("")
            print("Emotion Table")
            print("=============")
            # emotion table will insert blank lines before each section, so skip it here.
            avb.print_emotion_table()
            print("")

    def _print_image_table(avb: Binary) -> None:
        if show_images:
            print("")
            print("Image Table")
            print("===========")
            print("")
            avb.print_image_table()
            print("")

    def _info(avb: Binary) -> None:

        _print_summary_table(avb)
        _print_emotion_table(avb)
        try:
            avb.read_images(evaluate=True)
        except ExceptionGroup as exc_group:
            if show_images:
                # The image table will show all of these errors in a
                # suitable way for human consumption, avoid re-raising
                # them to be handled in the general case. Do, however,
                # log the exception info to the debugging stream.
                for exc in exc_group.exceptions:  # pylint: disable=not-an-iterable
                    _LOG.debug("Muffled error from read_images():", exc_info=exc)
                msg = f"There were {len(exc_group.exceptions)} problems parsing images."
                # debug handler will show full gore in the stack trace,
                # but full gore is already logged, so skip it.
                # in non-debug mode, we don't want the gore anyway.
                raise SystemExit(msg) from None
            raise
        finally:
            # read_images() may raise an Exception (or group), but we
            # still want to print the image table if we succeeded in
            # making one, which will contain error summaries based on
            # those exceptions in its output.
            _print_image_table(avb)

        if strict:
            try:
                avb.strictness_check()
                print("\nStrictness check: ✅")
            except Exception:
                print("\nStrictness check: ❌")
                print(_STRICT_MSG)
                raise

    with capture_warnings("cici") as warnings:
        with open_file(filename, parse_images=False, autorepair=False) as avb:
            try:
                _info(avb)
            finally:
                if show_warnings and warnings:
                    print("Warnings:")
                    for msg in warnings:
                        print(f"  - {msg}")
                    print("")


def register(subparsers: Any) -> None:
    subparser = subparsers.add_parser("info", help="Show information about AVB/BGB files")

    subparser.add_argument(
        "-e",
        "--emotions",
        action="store_true",
        help="Show pose/emotion/gesture information",
    )

    subparser.add_argument(
        "-i",
        "--images",
        action="store_true",
        help="Show image table information",
    )

    subparser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Verbose output: show emotion and image tables (implies -e, -i)",
    )

    subparser.add_argument(
        "-s",
        "--strict",
        action="store_true",
        help="Show pedantic warnings.",
    )

    subparser.add_argument(
        "--hide-warnings",
        action="store_true",
        help="Don't show warnings for non-fatal problems with the file.",
    )
    add_common_arguments(subparser)
    subparser.add_argument("filename")


def dispatch(args: argparse.Namespace) -> None:
    info(
        args.filename,
        args.strict,
        not args.hide_warnings,
        show_emotions=args.emotions or args.verbose,
        show_images=args.images or args.verbose,
    )



"""
Hello!
"""

from contextlib import contextmanager
import logging
import sys
from typing import (
    Dict,
    Iterator,
    List,
    Literal,
    Sequence,
)

import colorama
import termcolor


def multi_exception_shim(
    exceptions: Sequence[Exception], logger: logging.Logger, emsg: str
) -> None:
    if not exceptions:
        return

    if len(exceptions) == 1:
        raise exceptions[0]

    if sys.version_info >= (3, 11):
        raise ExceptionGroup(emsg, exceptions)

    # For older versions of Python, just raise the first Exception.

    for exc in exceptions[1:]:
        logger.error(f"Exception: {exc_summary(exc)}")
        logger.debug("Detail: ", exc_info=exc)

    raise exceptions[0]


def exc_name(exc: BaseException) -> str:
    name = type(exc).__qualname__
    smod = type(exc).__module__
    if smod not in ("__main__", "builtins"):
        name = smod + "." + name
    return name


def exc_summary(exc: BaseException) -> str:
    ret = str(exc)
    if ret:
        return f"{exc_name(exc)}: {ret}"
    return exc_name(exc)


Color = Literal[
    "black",
    "grey",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "light_grey",
    "dark_grey",
    "light_red",
    "light_green",
    "light_yellow",
    "light_blue",
    "light_magenta",
    "light_cyan",
    "white",
]


class ColorFormatter(logging.Formatter):
    colmap: Dict[int, Color] = {
        logging.INFO: "light_grey",
        logging.DEBUG: "dark_grey",
        logging.WARNING: "yellow",
        logging.ERROR: "red",
    }

    def format(self, record: logging.LogRecord) -> str:
        msg = super().format(record)
        if record.levelno in self.colmap:
            msg = termcolor.colored(msg, self.colmap[record.levelno])
        return msg


class WarningHandler(logging.Handler):
    def __init__(self, logger: logging.Logger) -> None:
        self.messages: List[str] = []
        self.logger = logger
        self.propagate = logger.propagate
        self.force = logger.getEffectiveLevel() <= logging.DEBUG

        # Force this handler to always "handle" events ...
        super().__init__(logging.NOTSET)

    def handle(self, record: logging.LogRecord) -> bool:
        propagate = self.propagate and (self.force or (record.levelno != logging.WARNING))
        if propagate:
            if parent := self.logger.parent:
                parent.callHandlers(record)

        if record.levelno == logging.WARNING:
            # only *actually* handle warnings.
            return super().handle(record)
        return False

    def emit(self, record: logging.LogRecord) -> None:
        msg = self.format(record)
        self.messages.append(msg)


@contextmanager
def capture_warnings(node: str) -> Iterator[List[str]]:
    """
    Simple context manager that captures logging at a given level.

    Messages will cease to propagate while in the context block; they
    will resume propagating (if they were before) upon exit from the
    block. Any messages captured during the block will be in the List
    yielded during the block.

    Example::
        with capture_logging('cici') as messages:
            avb = Binary('file.avb')

        for message in messages:
            print(message)
    """
    logger = logging.getLogger(node)
    propagate = logger.propagate

    # This weird little guy stores warnings for later, and also will
    # *conditionally* propagate only non-warning LogRecords, which
    # lets us fully capture warnings.
    handler = WarningHandler(logger)

    try:
        logger.propagate = False
        logger.addHandler(handler)
        yield handler.messages
    finally:
        logger.propagate = propagate
        logger.removeHandler(handler)


def setup_logging(debug: bool = False) -> None:
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.WARNING,
    )

    fmt = "%(levelname)s: %(message)s" if debug else "%(message)s"

    rootlogger = logging.getLogger()
    for handler in rootlogger.handlers:
        if isinstance(handler, logging.StreamHandler):
            handler.setFormatter(ColorFormatter(fmt))
            break

    colorama.init()



import argparse
from contextlib import contextmanager
import logging
import sys
from typing import Iterator

from cici.binary import Binary
from cici.util import exc_summary


def add_common_arguments(subparser: argparse.ArgumentParser) -> None:
    subparser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Show full debugging information and stack traces.",
    )


@contextmanager
def open_file(
    filename: str, parse_images: bool = True, autorepair: bool = False
) -> Iterator[Binary]:
    with open(filename, "rb") as file:
        try:
            avb = Binary(file, parse_images=False, autorepair=autorepair)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            print(f"Unable to parse file: {exc_summary(exc)}", file=sys.stderr)
            logging.debug("Unable to parse file", exc_info=True)
            sys.exit(1)

        if parse_images:
            try:
                avb.read_images()
            except Exception as exc:  # pylint: disable=broad-exception-caught
                print(f"Unable to parse images: {exc_summary(exc)}", file=sys.stderr)
                logging.debug("Unable to parse images", exc_info=True)
                sys.exit(1)

        yield avb



"""
The Emotion chip that Data was looking for this whole time

Unfortunately it was written in Python, so it tended to go a little
haywire.

This module implements the `OldEmotionHeader` class for v2.1 files,
and `EmotionHeader` for v2.5 files. Both classes inherit from the
abstract `EmotionHeaderBase` class which provides common functionality
for both.

Each one of these objects describes one entry in the
pose/gesture/emotion table; meaning it describes either one singular
pose, or for Microsoft-published avatar files, either one face or one
body.
"""

import enum
import logging
import struct
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

from PIL import ImageChops
from PIL.Image import Image

from .bitmap import PaletteType
from .imgmeta import ImageMetadata
from .table import IMG_TABLE


if TYPE_CHECKING:
    from .binary import Binary
    from .bitmap import Bitmap


_LOG = logging.getLogger(__name__)


class Emotion(enum.IntEnum):
    HAPPY = 1
    COY = 2
    BORED = 3
    SCARED = 4
    SAD = 5
    ANGRY = 6
    SHOUT = 7
    LAUGH = 8
    NEUTRAL = 9
    WAVE = 10
    POINT_TO_OTHER = 11
    POINT_TO_SELF = 12
    # The below emotions aren't accessible via ccedit nor cchat
    DOUBLEPOINT = 13  # mscc 2.5's LANCE
    SHRUG = 14  # mscc 2.5's DAN, SUSAN; and many user-edited files
    _3QRWALK = 15  # mscc 2.5's LANCE
    SIDEWALK = 16  # Not observed in the wild at all
    _3QFWALK = 17  # mscc 2.5's LANCE

    def __str__(self) -> str:
        return self.name.lstrip("_")


class EmotionHeaderBase:
    # pylint: disable=too-many-public-methods

    # okay like. theoretically yes this object is kind of doing too many things but
    # I am bad at OO design and I don't feel like refactoring everything right now, and like.
    # stop judging me, pylint!!!

    def __init__(self, parent: "Binary", field_type: int, n: int) -> None:
        # Metadata
        self._parent = parent
        self.field_type = field_type
        self.n = n

        # Cache: Although imgmeta caches raw images, *these* images
        # are potentially lightly transformed from the raw images in
        # the case of v2.5 avatars where mask + halo data is
        # multiplexed into a single image; also in the rarer case that
        # the mask/halo data is not correctly sized or contains
        # non-monochrome data; see _maskify().
        self._img1: Optional[Image] = None
        self._img2: Optional[Image] = None
        self._img3: Optional[Image] = None

        #####################
        # Actual field data #
        #####################

        # contains offsets, an unknown byte, and an image decoding flag.
        # The unknown byte/flag are only defined for v2.5 files.
        self._imgmeta = [
            ImageMetadata(parent, self.kind()),
            ImageMetadata(parent, "MASK"),
            ImageMetadata(parent, "HALO"),
        ]

        self.emotion: Union[int, Emotion] = 0
        self.intensity = 0
        self.coordinates_1 = (0, 0)  # Can be face center or junction point, depending.

        # Only for 0x04/0x0A (Faces)
        self.coordinates_2 = (0, 0)  # (Junction offset, when defined.)
        self.coordinates_3 = (0, 0)  # (Face center, when defined.)

    @staticmethod
    def describe_field(field_type: int) -> str:
        if field_type not in (0x04, 0x05, 0x09, 0x0A, 0x0B, 0x0C):
            raise ValueError("Unrecognized field type")

        noun = ver = ""
        if field_type in (0x04, 0x0A):
            noun = "Faces"
        if field_type in (0x05, 0x0B):
            noun = "Bodies"
        if field_type in (0x09, 0x0C):
            noun = "Poses"
        if field_type in (0x04, 0x05, 0x09):
            ver = "v2.1"
        if field_type in (0x0A, 0x0B, 0x0C):
            ver = "v2.5"

        assert noun and ver
        return f"{noun} ({ver})"

    @staticmethod
    def get_klass(field_type: int) -> Type["EmotionHeaderBase"]:
        if field_type in (0x04, 0x05, 0x09):
            return OldEmotionHeader
        assert field_type in (0x0A, 0x0B, 0x0C)
        return EmotionHeader

    def _warn(self, msg: str) -> None:
        prefix = f"Emotion {self.field_type:02X}/#{self.n:02d} ({str(self.emotion)}): "
        msg = prefix + msg
        _LOG.warning(msg)

    @property
    def face_coordinates(self) -> Optional[Tuple[int, int]]:
        """
        Return the facial center coordinates, if they exist.

        Faces and poses have facial coordinates, but bodies do
        not. These coordinates are stored in different locations in
        the serialized struct depending on what type of entry this is.

        Faces use the third set of coordinates, poses use the first
        set.
        """
        if self.face_data():
            return self.coordinates_3
        if self.body_data():
            return None
        return self.coordinates_1

    @property
    def junction_point(self) -> Optional[Tuple[int, int]]:
        """
        Return the face/body junction point coordinates.

        Faces and Bodies have junction coordinates, poses do not.
        """
        if self.face_data() or self.body_data():
            return self.coordinates_1
        return None

    @property
    def junction_offset(self) -> Optional[Tuple[int, int]]:
        """
        Return the face junction point offset.

        Only faces have this offset.
        """
        if self.face_data():
            return self.coordinates_2
        return None

    def face_data(self) -> bool:
        """
        True if this data represents a FACE.
        """
        return self.field_type in (0x04, 0x0A)

    def body_data(self) -> bool:
        """
        True if this data represents a BODY.
        """
        return self.field_type in (0x05, 0x0B)

    def kind(self) -> str:
        """
        What kind of data is this?

        :return: "FACE", "BODY", or "POSE".
        """
        if self.face_data():
            return "FACE"
        if self.body_data():
            return "BODY"
        return "POSE"

    def intensity_wire(self) -> int:
        """
        This function computes the intensity of the emotion for
        the MSCC control message wire protocol.

        Each message sent from Comic Chat sends an intensity for the
        emotion, ranging from 0-10 inclusive. Notably, specialty poses
        like Wave, Neutral, Point_to_self, and Point_to_other always
        use 0, but this value can also be used by very "mild"
        expressions/gestures.

        (NOTE: Non-specialty emotions such as Happy/Sad etc cannot use
        a literal intensity of 0x00, that will crash Comic Chat!)

        It is *closely related to*, but *different* from the intensity
        slider visible in CCEdit, which follows a maddening, subtly
        different algorithm.

        This function *is* defined in CChat for all values from 0x00
        to 0xFF, even though only certain values appear in avatars
        made with official editors.

        The table below shows each wire intensity alongside the range
        of actual emotional intensity values that produce that value,
        and the total number of values that map to that single wire
        intensity.

        0:  [0x00 - 0x19]  26
        1:  [0x1A - 0x32]  25
        2:  [0x33 - 0x4C]  26
        3:  [0x4D - 0x65]  25
        4:  [0x66 - 0x7F]  26
        5:  [0x80 - 0x98]  25
        6:  [0x99 - 0xB2]  26
        7:  [0xB3 - 0xCB]  25
        8:  [0xCC - 0xE5]  26
        9:  [0xE6 - 0xFE]  25
        ::  [    0xFF   ]   1

        It's worth noting that the intensity 0xFF produces the wire
        value ":", which is "0" + 10. This function returns an integer
        and not a char, however.
        """
        return int((self.intensity * 10) / 255)

    def intensity_v55(self) -> int:
        """
        Return the CCEdit v55 slider intensity for this emotion.

        Returns a number from 0 to 10, with 0 meaning "low" emotion
        and 10 meaning "high" emotion. This is backwards from the v66
        convention.

        v55 used "1" for low intensity and "10" for maximum
        intensity. However, the literal byte values it chose for the
        emotion intensity had the unfortunate property that they were
        not unique for the wire protocol and had several collisions:

        idx   EI      wire
        -----------------
         0: 0x00  =>  0         (specialty poses)
         1: 0x19  =>  0         (emotion wheel: near the center, low)
         2: 0x33  =>  2
         3: 0x4C  =>  2
         4: 0x66  =>  4
         5: 0x7F  =>  4
         6: 0x99  =>  6
         7: 0xB2  =>  6
         8: 0xCC  =>  8
         9: 0xE5  =>  8
        10: 0xFF  =>  :         (emotion wheel: near the edge, high)

        The mapping function for index to intensity here is:
        f(x) = int(255/10 * x)

        This function performs the inverse and returns the intensity
        index (0-10) for a given emotion intensity.

        For "unusual" intensity values (created by a hex editor or v66
        of CCedit), the numerical indexes returned by this function
        are only hypothetical, since there is no legitimate way to
        produce those values in v55.

        The inverse intensity mapping for this function is:

        [0x00 - 0x18]: 0
        [0x19 - 0x32]: 1
        [0x33 - 0x4B]: 2
        [0x4C - 0x65]: 3
        [0x66 - 0x7E]: 4
        [0x7F - 0x98]: 5
        [0x99 - 0xB1]: 6
        [0xB2 - 0xCB]: 7
        [0xCC - 0xE4]: 8
        [0xE5 - 0xFE]: 9
        [   0xFF    ]: 10

        """
        return int((self.intensity + 0.5) / 25.5)

    def intensity_v66(self) -> int:
        """
        Return the CCEdit v66 slider intensity for this emotion.

        Returns a number from 0 to 10, with 10 meaning "high" emotion
        and 0 meaning "low" emotion. This is backwards from the v55
        convention.

        v66 used "10" for low intensity and "1" for maximum
        intensity. It is inverted from the v55 convention; but avoids
        collisions when mapping to the wire intensity.

        idx   EI      wire
        -----------------
         0: 0x00  =>  0         (specialty poses)
        10: 0x01  =>  0         (emotion wheel: near the center, low)
         9: 0x1B  =>  1
         8: 0x34  =>  2
         7: 0x4E  =>  3
         6: 0x67  =>  4
         5: 0x81  =>  5
         4: 0x9A  =>  6
         3: 0xB4  =>  7
         2: 0xCD  =>  8
         1: 0xE7  =>  9         (emotion wheel: near the edge, high)

        This mapping from index to intensity is:
        f(x) = (256 - int(255/10 * x)) % 256

        Notably, v66 cannot produce the intensity value 0xFF nor the
        wire intensity 10. The slider value of "0" for specialty
        poses is special-cased to the intensity value 0x00.

        Like the v55 function, intensity values outside of the
        well-defined values are hypothetical, since there was no way
        to produce those values with the official v66 editor.

        The inverse intensity mapping for this function is:

        [   0x00    ]: 0
        [0x01 - 0x1A]: 10
        [0x1B - 0x33]: 9
        [0x34 - 0x4D]: 8
        [0x4E - 0x66]: 7
        [0x67 - 0x80]: 6
        [0x81 - 0x99]: 5
        [0x9A - 0xB3]: 4
        [0xB4 - 0xCC]: 3
        [0xCD - 0xE6]: 2
        [0xE7 - 0xFF]: 1

        """
        if self.intensity == 0x00:
            return 0
        return 10 - int((self.intensity - 1) / 25.5)

    def describe(self) -> str:
        # Specialty poses do not use/advertise an intensity subscript
        if self.emotion >= Emotion.NEUTRAL:
            subscript = ""
        else:
            # Use the wire intensity because it is consistent across
            # v55 and v66 made avatars; and it actually represents how
            # the gestures are communicated over the IRC protocol.
            subscript = f"[{self.intensity_wire()}]"

        return f"#{self.n:02d} {str(self.emotion)}{subscript}"

    def unstable_hash(self) -> int:
        return hash(self._imgmeta[0].offset)

    def read(self, file: BinaryIO) -> None:
        data = struct.unpack("<IIIh", file.read(14))
        for i in range(3):
            self._imgmeta[i].offset = data[i]

        # NB: It's not safe to use _warn until self.emotion is set.
        try:
            self.emotion = Emotion(data[3])
        except ValueError:
            self.emotion = data[3]
            self._warn(f"Found unknown emotion #{self.emotion}")
        if self.emotion >= 13:
            self._warn(f"Emotion ({str(self.emotion)}) is inaccessible in CChat")

        if not any(self._imgmeta[i].offset for i in range(3)):
            self._warn("Emotion has zero attached images. That's probably not right.")

        # Note: x/y appear to be stored as 2 byte integers, but the
        # HOB is *almost always zero*. Need to investigate to see if
        # there's a CCedit limitation or a CChat one.
        self.intensity, x, y = struct.unpack("<Bhh", file.read(5))
        self.coordinates_1 = (x, y)

        if self.face_data():
            # HOB on these is sometimes used; it's definitely signed.
            self.coordinates_2 = cast(Tuple[int, int], struct.unpack("<hh", file.read(4)))
            # HOB are (almost always) always zero again.
            self.coordinates_3 = cast(Tuple[int, int], struct.unpack("<hh", file.read(4)))

        # NB: It's not safe to use describe() until intensity is set.
        for imgmeta in self._imgmeta:
            imgmeta.context = self.describe()

    def size(self) -> int:
        # NB: This represents an abstract size,
        # v2.1 and v2.5 records are both bigger in practice.
        return 19 + (8 if self.face_data() else 0)

    def __bytes__(self) -> bytes:
        data = struct.pack(
            "<IIIhBhh",
            self._imgmeta[0].offset,
            self._imgmeta[1].offset,
            self._imgmeta[2].offset,
            int(self.emotion),
            self.intensity,
            self.coordinates_1[0],
            self.coordinates_1[1],
        )
        if self.face_data():
            data += struct.pack(
                "<hhhh",
                self.coordinates_2[0],
                self.coordinates_2[1],
                self.coordinates_3[0],
                self.coordinates_3[1],
            )
        return data

    def bitmap(self, index: int) -> Optional["Bitmap"]:
        """
        Retrieve a raw bitmap associated with this emotion/pose.

        :param index: Valid values are 0-2, inclusive.
        :return: A Bitmap object, if the image at that index exists.
        """
        return self._imgmeta[index].bmp

    def update_offset(self, index: int, offset: int) -> None:
        """
        Update the file offset for a specified bitmap.

        This is used by the Binary writer when it re-arranges the
        image table to write back out to file. You shouldn't need to
        use this function under normal circumstances.

        :param index: Which bitmap's offset to update, [0-2]
        :param offset: Virtual (unrebased) offset to bitmap; uint32_t.
        """
        self._imgmeta[index].offset = offset

    def raw_image(self, index: int) -> Optional[Image]:
        """
        Retrieve a raw image associated with this emotion/pose.

        :param index: Valid values are 0-2, inclusive.
        :return: a PIL Image object, if the image at that index exists.
        """
        return self._imgmeta[index].img

    def base_img(self) -> Image:
        """
        Retrieve the base image for this emotion/pose.

        :return: a PIL Image object.
        """
        if self._img1:
            return self._img1
        self._img1 = self.raw_image(0)
        assert self._img1
        return self._img1

    def _maskify(self, img: Image, kind: str) -> Image:
        # Turns mask data into a mask we can use with PIL.
        base = self.bitmap(0)
        assert base is not None

        if len(img.getcolors()) > 2:
            self._warn(f"{kind} uses more than 2 colors")

        base_img_lum = base.get_img().convert("L")
        mask_img_lum = img.convert("L")

        if mask_img_lum.size < base.dimensions:
            self._warn(f"{kind} is smaller than the base image and will be upscaled")
            mask_img_lum = mask_img_lum.resize(base.dimensions)

        # Normally, any white pixels in the mask correspond to white
        # pixels in the base image. These pixels become fully
        # transparent when composited onto a background. If there are
        # white pixels in the mask that happen to overlay non-white
        # pixels (this is extremely rare in avatars in circulation!)
        # this may induce partial transparency. To emulate this,
        # multiply the mask against the base image. In most cases,
        # this will have no effect at all; in rare cases, this will
        # "grow" the mask to include any sections of the base image
        # that have non-white pixels.
        img = ImageChops.multiply(base_img_lum, mask_img_lum)
        if list(img.getdata()) != list(mask_img_lum.getdata()):
            self._warn(f"{kind} selects non-white pixels and has been modified")

        img = ImageChops.invert(img)
        return img

    def halo_img(self) -> Image:
        # "Halo" mask - this is a looser fitting mask that adds the "halo" border.
        if self._img3:
            return self._img3
        img = self.raw_image(2)
        if img is None:
            raise RuntimeError("No halo image defined")
        self._img3 = self._maskify(img, "halo")
        return self._img3

    def mask_img(self) -> Image:
        # "Tight" mask - this is used for compositing faces onto bodies.
        # (Or in rare cases, bodies onto faces!)
        if self._img2:
            return self._img2
        img = self.raw_image(1)
        if img is None:
            raise RuntimeError("No mask image defined")
        self._img2 = self._maskify(img, "mask")
        return self._img2

    def render(self, transparency: str = "halo") -> Image:
        """
        Used for getting different composite images.

        "halo" - use the halo mask to composite this image.
        "mask-preferred" - use the tight mask to composite this image, but fall back to the halo.
        "mask-none" - use the tight mask if possible, fall back to no transparency.
        "none" - do not use any transparency mask.
        """
        if transparency not in ("halo", "mask-preferred", "mask-none", "none"):
            raise ValueError(f"unrecognized transparency argument {transparency}")

        img = self.base_img()
        img = img.convert("RGBA")
        alpha = None

        if transparency in ("mask-preferred", "mask-none"):
            try:
                alpha = self.mask_img()
            except RuntimeError:
                pass

        if transparency in ("mask-preferred", "halo"):
            if alpha is None:
                alpha = self.halo_img()

        if alpha:
            img.putalpha(alpha)
        return img

    def fingerprint(self) -> Tuple[Any, ...]:
        """
        Return fingerprint data for this emotion.

        This data includes data characteristic of this emotion entry,
        used ultimately for computing the "header fingerprint" of the
        entire file. It cannot detect changes in the pose data itself,
        but as a result is tolerant of offset differences. This is
        ultimately a heuristic for determining when two different
        avatars have different metadata, but the "same poses".

        Modifying this function will change the header fingerprint
        algorithm and output, so do not change it without very good
        reason.
        """
        return (
            int(self.emotion),
            self.intensity,
            bool(self._imgmeta[0]),
            bool(self._imgmeta[1]),
            bool(self._imgmeta[2]),
            self.coordinates_1,
            self.coordinates_2,
            self.coordinates_3,
            self._imgmeta[0].flag or 0,
            self._imgmeta[1].flag or 0,
            self._imgmeta[2].flag or 0,
        )

    @classmethod
    def table_header(cls, combined: bool = True) -> str:
        ret = "## kind offset_1 offset_2 offset_3 Emotion        EI xy_coord_1"
        if not combined:
            ret += " xy_coord_2 xy_coord_3"
        return ret

    def _summary_fields(self) -> List[str]:
        # Used for table output.
        return [
            f"{self.n:02d}",
            self.kind(),
            f"{self._imgmeta[0].offset:08X}",
            f"{self._imgmeta[1].offset:08X}",
            f"{self._imgmeta[2].offset:08X}",
            f"{str(self.emotion):14}",
            f"{self.intensity:02X}",
        ]

    def _summary_fields_suffix(self) -> List[str]:
        # Used for table output.
        ret = [
            f"({self.coordinates_1[0]:3d}, {self.coordinates_1[1]:3d})",
        ]
        if self.face_data():
            ret.append(f"({self.coordinates_2[0]:3d}, {self.coordinates_2[1]:3d})")
            ret.append(f"({self.coordinates_3[0]:3d}, {self.coordinates_3[1]:3d})")
        elif self.body_data():
            ret.append("    --    ")
            ret.append("    --    ")
        return ret

    def get_summary_row(self) -> str:
        fields = self._summary_fields() + self._summary_fields_suffix()
        return " ".join(fields)

    def get_img_table_rows(self, evaluate: bool = False) -> List[Tuple[str, Optional[Exception]]]:
        return [imgmeta.img_table_row(evaluate) for imgmeta in self._imgmeta if imgmeta]

    def strictness_check(self) -> None:
        # The order in which the emotions are stored appears to follow
        # a priority order:
        #
        # Neutral > Laugh > Bored > Angry > Happy > Scared > Shout >
        # Sad > Coy > Point to Self > Point to Other > Wave
        #
        # Maybe I'll check this (someday)

        # Although any intensity is generally valid, only a select few
        # values were produced by official editors and have generally
        # reliable behavior. Warn for unusual intensity values.
        assert self.intensity in (
            0x00,
            0x01,
            0x1B,
            0x34,
            0x4E,
            0x67,
            0x81,
            0x9A,
            0xB4,
            0xCD,
            0xE7,
            0x19,
            0x33,
            0x4C,
            0x66,
            0x7F,
            0x99,
            0xB2,
            0xCC,
            0xE5,
            0xFF,
        ), f"Unusual emotional intensity (EI) value 0x{self.intensity:02X}"

        # Check the face center coordinates. The official editor was
        # buggy and could produce bizarre values in some
        # circumstances, and hex-edited avatars often feature bizarre
        # coordinates.

        # In my own testing, the face locations I calculated as being
        # at 0x0F would glitch out and select a bigger coordinate.  I
        # couldn't select any coordinate lower than this.

        # When transparent margins are present, however, you can
        # select a location overlapping with or purely in the margin
        # to get lower values. So low it can underflow back up to
        # 0xFF!

        # Try to guess a canvas size for this character.
        # This can't be known with certainty; this is only the largest we can prove.
        # In the case that some calculation below fails to make sense with this palette size,
        # we may fall back to guessing the default 154x224, but it's not possible to know
        # for sure what the real canvas size actually was.
        w, h = self._parent.guess_canvas_size()
        bmp = self.bitmap(0)
        assert bmp is not None

        # NB: Bodies don't have face coordinates O:-)
        if self.face_coordinates is not None and self.face_coordinates[0] >= w:
            if (w - bmp.dib.width) > (256 - self.face_coordinates[0] + 16):
                # The value is too big, but it might be explained as
                # underflow and we are centered in the left margin -
                # the value underflowed due to the crop.
                self._warn(f"face-x appears to have underflown ({w}x{h} canvas)")
            elif (154 - bmp.dib.width) > (256 - self.face_coordinates[0] + 16):
                # Might still be underflow if we assume the default 154x224 canvas.
                self._warn("face-x appears to have underflown (assuming 154x224 canvas)")
            elif self.face_coordinates[0] < 0x90:
                # 0x89, 137, is the highest value you can get using
                # the standard template size of 154 x 224. In this
                # case, we assume the point is deep into the margin on the right.
                self._warn("face-x appears to be in the right margin.")
            else:
                self._warn(
                    f"face-x {self.face_coordinates[0]} exceeds canvas width ({w})"
                    " and standard canvas width (154) and cannot be explained by underflow"
                )

        if self.face_coordinates is not None and self.face_coordinates[1] >= h:
            if (h - bmp.dib.height) > (256 - self.face_coordinates[1] + 16):
                # Value is too big, but it might be underflow again.
                # We might be very high up in the top margin that got
                # cropped off.
                self._warn(f"face-y appears to have underflown ({w}x{h} canvas)")
            elif (224 - bmp.dib.height) > (256 - self.face_coordinates[1] + 16):
                self._warn("face-y appears to have underflown (assuming 154x224 canvas)")
            elif self.face_coordinates[1] < 0xD0:
                # 0xCF, 207, is the highest value you can get using
                # the standard template size of 154 x 224. In this
                # case, we assume the point is deep into the margin on
                # the bottom.
                self._warn("face-y appears to be in the bottom margin.")
            else:
                self._warn(
                    f"face-y {self.face_coordinates[1]}"
                    f" exceeds canvas height ({h})"
                    " and standard canvas height (224) and cannot be explained by underflow"
                )

        # These coordinates never seem to use the HOB.
        # assert 0 <= self.coordinates_1[0] <= 255
        # Re-Man uses the HOB for a single pose! It's just uncommon. Seems to work well.
        assert 0 <= self.coordinates_1[1] <= 255
        if self.face_data():
            assert 0 <= self.coordinates_3[0] <= 255
            assert 0 <= self.coordinates_3[1] <= 255


class OldEmotionHeader(EmotionHeaderBase):
    # This represents Poses, Faces and Bodies for v2.1 files.

    def __init__(self, parent: "Binary", field_type: int, n: int) -> None:
        super().__init__(parent, field_type, n)
        self.reserved_16: bytes = b"\x00" * 16

    def read(self, file: BinaryIO) -> None:
        super().read(file)
        self.reserved_16 = file.read(16)

    def __bytes__(self) -> bytes:
        return super().__bytes__() + self.reserved_16

    def size(self) -> int:
        return super().size() + 16

    def strictness_check(self) -> None:
        super().strictness_check()
        assert self.reserved_16 == b"\x00" * 16

        assert self._imgmeta[0]
        bmp = self.bitmap(0)
        assert bmp is not None

        # This is just a nice way to see the spread of possible image
        # types I've seen. There's no inherent reason why ones not
        # present here wouldn't work, but it's nice to have a
        # reference for what's out there.

        info = bmp.characterize()
        assert info in IMG_TABLE["v55"][0], f"unexpected image type for image: {info}"

        # Random factoid: 1bpp images that store their palette as
        # (black, white) are mode='1' to PIL, whereas (white, black)
        # become 1bpp 'P' type.  Silly, yet true!

        if self._parent.composition == 0x05:
            # For files that render face-on-body (most dynamic avatars),
            # we require the presence of a tight mask which is for
            # v2.1 files always stored in the offset_2 slot.
            if self.face_data():
                assert self._imgmeta[1], "offset_2 must be set for facial data"
            if self.body_data():
                # Well, it's more that it's extraneous and I've never seen it happen.
                # I don't think it actually hurts anything.
                assert not self._imgmeta[1], "offset_2 shouldn't be set for body data"
            if not (self.face_data() or self.body_data()):
                # Similarly, I doubt this is harmful - I've just never seen it.
                assert not self._imgmeta[1], "offset_2 shouldn't be set for pose data"
        elif self._parent.composition == 0x02:
            # These are files that render body-on-face (SUSAN only).
            if self.face_data() and self._imgmeta[1]:
                self._warn("Face Mask not expected to be set when composition == 0x02")
            if self.body_data():
                assert self._imgmeta[
                    1
                ], "offset_2 must be set for body data for composition == 0x02"

        if bmp := self.bitmap(1):
            info = bmp.characterize()
            assert info in IMG_TABLE["v55"][1], f"unexpected image type for mask: {info}"
            img = bmp.get_img()
            img = img.convert("RGB")
            colors = img.getcolors()
            assert len(colors) <= 2, f"ofs2/mask uses more than 2 colors ({len(colors)}) (v2.1)"
            colorset = set(pair[1] for pair in colors)
            assert colorset <= {
                (255, 255, 255),
                (0, 0, 0),
            }, f"ofs2/mask colorset {colorset}"

        # offset_3, the halo image, must be defined!
        assert self._imgmeta[2], "missing halo mask"
        bmp = self.bitmap(2)
        assert bmp is not None
        info = bmp.characterize()
        assert info in IMG_TABLE["v55"][2], f"unexpected image type for halo: {info}"

        # Forego checking the halo colors here -- if it's more than 2 colors,
        # we'll emit a warning during normal parsing instead.
        # In practice, 99.9% of files use a halo image that's black and white.
        # Only a single file I observed didn't, and that file had only a single
        # pixel that was (255, 255, 254) that violates this assumption.


class EmotionHeader(EmotionHeaderBase):
    # This represents Poses, Faces and Bodies for v2.5 files.

    def read(self, file: BinaryIO) -> None:
        super().read(file)

        data = struct.unpack("<BBB", file.read(3))
        for i in range(3):
            self._imgmeta[i].unk = data[i]

        data = struct.unpack("<BBB", file.read(3))
        for i in range(3):
            self._imgmeta[i].flag = data[i]

    def __bytes__(self) -> bytes:
        return super().__bytes__() + struct.pack(
            "<BBBBBB",
            self._imgmeta[0].unk,
            self._imgmeta[1].unk,
            self._imgmeta[2].unk,
            self._imgmeta[0].flag,
            self._imgmeta[1].flag,
            self._imgmeta[2].flag,
        )

    def size(self) -> int:
        return super().size() + 6

    def _split_2bpp_image(self) -> None:
        """
        This method "splits" a 2bpp image that uses arbitrary
        colors (as explained in Bitmap.py) into three separate black
        and white component images.
        """
        img = self.raw_image(0)
        assert img is not None
        self._img1 = img.getchannel("R")
        self._img2 = self._maskify(img.getchannel("G"), "mask")
        self._img3 = self._maskify(img.getchannel("B"), "halo")

    def base_img(self) -> Image:
        if self._img1:
            return self._img1
        if self._imgmeta[0].flag == 0x04:
            self._split_2bpp_image()
            assert self._img1 is not None
            return self._img1
        return super().base_img()

    def mask_img(self) -> Image:
        if self._img2:
            return self._img2
        if self._imgmeta[0].flag == 0x04:
            self._split_2bpp_image()
            assert self._img2 is not None
            return self._img2
        return super().mask_img()

    def halo_img(self) -> Image:
        if self._img3:
            return self._img3
        if self._imgmeta[0].flag == 0x04:
            # Halo is encoded in image1
            self._split_2bpp_image()
            assert self._img3 is not None
            return self._img3
        if self._imgmeta[1].flag == 0x05:
            # There's *no* halo, so use the tight mask from img2.
            self._img3 = self.mask_img()
            return self._img3
        return super().halo_img()

    @classmethod
    def table_header(cls, combined: bool = True) -> str:
        ret = "## kind offset_1 offset_2 offset_3 Emotion        EI F1 F2 F3 xy_coord_1"
        if not combined:
            ret += " xy_coord_2 xy_coord_3"
        return ret

    def _summary_fields(self) -> List[str]:
        return super()._summary_fields() + [
            f"{self._imgmeta[0].flag:02X}",
            f"{self._imgmeta[1].flag:02X}",
            f"{self._imgmeta[2].flag:02X}",
        ]

    def strictness_check(self) -> None:
        super().strictness_check()

        assert self._imgmeta[0].unk == 0x01
        assert self._imgmeta[1].unk == 0x01
        assert self._imgmeta[2].unk == 0x01
        # Redundant with checks below, but nice to see in one place:
        assert self._imgmeta[0].flag in (0x02, 0x04)
        assert self._imgmeta[1].flag in (0x03, 0x05)
        assert self._imgmeta[2].flag == 0x03

        # All poses and bodies have a halo.
        # All faces have a halo and a mask.
        # But, where are they hiding? O:-)
        if self._imgmeta[0].flag == 0x04:
            # offset_1 has the implied 2bpp format.
            # It can be used for faces, bodies, or poses. (!)
            # In any event, it contains all the transparencies we need.
            assert not self._imgmeta[1]
            assert not self._imgmeta[2]
        elif self._imgmeta[0].flag == 0x02:
            # offset_1 has a normal kind of image, so we need to get
            # our transparencies from somewhere.  One or the other
            # should be defined, but not both.
            assert self._imgmeta[1] or self._imgmeta[2]
            assert not (self._imgmeta[1] and self._imgmeta[2])

        assert self._imgmeta[0]
        bmp = self.bitmap(0)
        assert bmp
        info = bmp.characterize((self._imgmeta[0].flag,))
        assert info in IMG_TABLE["v66"][0], f"unexpected image type for image: {info}"

        if bmp.palette_type() == PaletteType.IMPLICIT:
            # All this block really tests is that we got a 2bpp flag=0x04 image
            # that actually encodes one or more monochrome images and masks in one
            # image field.
            assert self._imgmeta[0].flag == 0x04
            img = bmp.get_img()
            colors = img.getcolors()
            assert (
                len(colors) <= 4
            ), f"ofs1 implied palette img doesn't have <=4 colors {len(colors)}"
            # PIL's typing extension package seems to lie about the type of getcolors();
            colorset = cast(Set[Tuple[int, int, int, int]], set(pair[1] for pair in colors))
            assert colorset <= {
                (255, 255, 255, 255),
                (255, 255, 0, 255),
                (255, 0, 0, 255),
                (0, 0, 0, 255),
            }, f"ofs1 colorset {colorset}"

            if self.kind() == "POSE":
                assert len(colors) <= 3, "pose colors"
                # This encodes the base image + halo.
                #
                # The way this gets encoded is by combining the halo
                # into the white foreground data; effectively
                # "eliminating" the dedicated "halo" channel. Therefore,
                # we have red (white foreground), black (black
                # foreground), and white (transparency) pixels only.
                assert colorset <= {
                    (255, 0, 0, 255),
                    (255, 255, 255, 255),
                    (0, 0, 0, 255),
                }, f"pose1 colorset {colorset}"
            else:
                if (self.kind() == "BODY" and self._parent.composition == 0x05) or (
                    self.kind() == "FACE" and self._parent.composition == 0x02
                ):
                    # This *also* encodes a base image + halo.

                    # The tight mask is almost never
                    # present. *sometimes*, though, a superfluous halo
                    # is included anyway -- SUSAN.AVB does this for
                    # her final two facial emotions.

                    # Like the POSE section above, the halo data is
                    # usually combined with the foreground data; in
                    # the case it isn't and we have standalone HALO
                    # pixels, we know we have superfluous tight mask
                    # data.
                    if (255, 255, 0, 255) in colorset:
                        self._warn("2bpp multi-channel image includes superfluous tight mask data.")

        if bmp := self.bitmap(1):
            info = bmp.characterize((self._imgmeta[1].flag,))
            assert info in IMG_TABLE["v66"][1], f"unexpected image type for ofs2/mask: {info}"
            img = bmp.get_img()
            colors = img.getcolors()
            assert len(colors) == 2, "ofs2/mask doesn't have exactly 2 colors"
            # As above, PIL's type extension is not correct about the type of getcolors()
            colorset = cast(Set[Tuple[int, int, int, int]], set(pair[1] for pair in colors))
            assert colorset <= {
                (0, 0, 0, 255),
                (255, 255, 255, 255),
            }, f"ofs2 colorset {colorset} emotion {self.n}"
            # offset_2 is very scant! This would be the "tight mask" for heads only;
            # but that mask is packed into offset_1 2bpp when possible instead, I think.
        else:
            # For some reason ...?
            assert self._imgmeta[1].flag == 0x03, "absent ofs2 should use 03 flag"

        if bmp := self.bitmap(2):
            info = bmp.characterize((self._imgmeta[2].flag,))
            assert info in IMG_TABLE["v66"][2], f"unexpected image type for ofs3/halo: {info}"

            img = bmp.get_img()
            colors = img.getcolors()
            assert len(colors) <= 2, "ofs3/halo uses more than 2 colors (v2.5)"
        else:
            assert self._imgmeta[2].flag == 0x03



from typing import (
    TYPE_CHECKING,
    List,
    Optional,
    Tuple,
)

from PIL import Image

from .bitmap import Bitmap
from .util import exc_summary


if TYPE_CHECKING:
    from .binary import Binary


class ImageMetadata:
    """
    Represents an image header entry as seen in a Comic Chat AVB/BGB file.

    :param parent: The parent Binary object, needed for actually
        retrieving the image data.
    :param kind: One of POSE, FACE, BODY, ICON or BKGD. Used for
        display in the image table summary, as well as conditional
        fixes for improperly stored images.

    :param offset: Virtual offset to image data.
    :param unk: Unknown header flag; always 0x01 in v2.5 files.
    :param flag: Image compression type for v2.5 files.
    """

    def __init__(
        self,
        parent: "Binary",
        kind: str,
        context: Optional[str] = None,
        *,
        offset: Optional[int] = None,
        unk: Optional[int] = None,
        flag: Optional[int] = None,
    ) -> None:
        self._parent = parent
        self._kind = kind
        self._bmp: Optional[Bitmap] = None
        self._img: Optional[Image.Image] = None
        self.context = context
        self.offset = offset
        self.unk = unk  # None (2.1) / 0x01 (2.5)
        self.flag = flag  # None (2.1) / 0x02 - 0x05 (2.5)

    def __bool__(self) -> bool:
        return bool(self.offset)

    @property
    def bmp(self) -> Optional[Bitmap]:
        if self._bmp:
            return self._bmp

        if not self.offset:
            return None

        self._bmp = self._parent.get_bmp(self.offset, self.flag or 0)
        return self._bmp

    @property
    def img(self) -> Optional[Image.Image]:
        if self._img:
            return self._img

        if self.bmp is None:
            return None

        # If this is an icon and the flag is 0x04, it's been
        # mistakenly processed as a 2bpp mask. We need to pretend
        # it's a 1bpp image instead and force it back to a
        # regular, monochromatic 1bpp icon.
        use_hack = self.flag == 0x04 and self._kind == "ICON"
        self._img = self.bmp.get_img(hack_2bpp_to_1bpp=use_hack)
        return self._img

    @staticmethod
    def img_table_hdr() -> List[str]:
        return [
            "  PARENT                 TYPE @vOFFSET @ACTUAL    WxH   BPP #COL Palette ",
            "  -----------------------------------------------------------------------",
        ]

    def img_table_row(self, evaluate: bool = False) -> Tuple[str, Optional[Exception]]:
        """
        Return a table row entry for this image metadata object.

        Because some image metadata entries may lead to corrupt or
        un-parseable entries, this method returns a tuple containing
        the row entry itself (filling as many fields as it can), and
        an Exception object, if any, with full details as to why the
        bmp could not be retrieved.

        NOTE: This method forces evaluation of the .bmp, but does not
        attempt to deserialize the image itself; for Comic Chat v2.5
        avatars, this means that there may be lingering issues with
        the image if e.g. the bitmap metadata is retrieved
        successfully, but there are issues with the zlib data that
        follows.
        """
        if not self.offset:
            return ("", None)

        real_offset = self._parent.get_offset(self.offset)
        ret = f"  {self.context:22s} {self._kind:4s} {self.offset:08X} {real_offset:08X}"

        try:
            bmp = self.bmp
            assert bmp is not None
        except Exception as exc:  # pylint: disable=broad-exception-caught

            # The caller is responsible for the exception data: we
            # want to provide a cohesive image data table, so we can't
            # stop for small errors. BaseExceptions are allowed to
            # halt the program, though.

            # NB: _parse_images will collect these exceptions and
            # log/raise them appropriately via multi_exception_shim,
            # so we don't need to log them here.

            exc.add_note(
                f"For {self.context} {self._kind} image; "
                f"vOffset=0x{self.offset:08X} offset=0x{real_offset:08X}"
            )

            ret += f"                           ❌ {exc_summary(exc)}"
            return (ret, exc)

        ret += (
            f" {bmp.dib.width:03d}x{bmp.dib.height:03d}"
            f" {bmp.dib.bits_per_pixel:2d}  {bmp.dib.n_colors:3d}  {bmp.palette_type():8s}"
        )

        if evaluate:
            try:
                _ = self.img
            except Exception as exc:  # pylint: disable=broad-exception-caught

                # Same exception-handling safety commentary as above.
                exc.add_note(
                    f"For {self.context} {self._kind} image; "
                    f"vOffset=0x{self.offset:08X} offset=0x{real_offset:08X}"
                )
                ret += f" ❌ {exc_summary(exc)}"
                return (ret, exc)

        return (ret, None)



"""
Bitmap Module

This provides the `Bitmap` and `CompressedBitmap` classes, which
provide support for reading the images stored in CChat 2.1 and 2.5
files.
"""

from contextlib import contextmanager
from enum import Enum
from io import BytesIO
import itertools
import logging
import math
import os
from pathlib import Path
import re
import struct
from typing import (
    Any,
    BinaryIO,
    Iterator,
    Optional,
    Tuple,
    Union,
)
import zlib

from PIL.Image import Image, frombytes
from PIL.Image import open as image_open


_LOG = logging.getLogger(__name__)


class PaletteType(Enum):
    # Standard Bitmap, no Palette
    NONE = 0
    # Standard Bitmap, with Palette
    EMBEDDED = 1
    # Compressed Bitmap, with compressed palette
    EXPLICIT = 2
    # Compressed Bitmap, with empty compressed palette
    EMPTY = 3
    # Compressed Bitmap; *without* compressed palette
    IMPLICIT = 4

    def __str__(self) -> str:
        return self.name.title()


class Bitmap:
    """
    Class for de/serializing standard .bmp files.

    Comic Chat 2.1 avatars store images as regular embedded BMP
    files. 2.5 files use a kind of modified BMP file that is handled
    separately in the `CompressedBitmap` subclass.

    BMP files have the following structure:
    0x00 - Bitmap file header  (14 bytes)
    0x0E - DIB Header          (40 bytes)
    0x36 - Palette, if present
         - Pixel data [immediately after Palette or DIB]

    For more info, see https://en.wikipedia.org/wiki/BMP_file_format#File_structure
    """

    _HEADER_SIZE = 14  # (includes magic through offset)

    def __init__(
        self,
        initial: Optional[Union[bytes, str, Path, "Bitmap", BinaryIO]] = None,
        autorepair: bool = False,
    ):
        # Binary data, in order;
        #
        # underscored fields have getters/setters instead that help
        # enforce consistency.

        self.header: bytes  # Covers the following:
        self.magic = b"BM"
        self._filesize = 0
        self.reserved_a = 0
        self.reserved_b = 0
        self._offset = 0

        self.dib = DIBHeader()
        self._palette = b""
        self._data = b""

        # Cache/Options

        self._img: Optional[Image] = None
        self._autorepair = autorepair
        self._invalidate = True

        file: BinaryIO
        if isinstance(initial, bytes):
            file = BytesIO(initial)
            self.read(file)
        elif isinstance(initial, (str, Path)):
            with open(initial, "rb") as file:
                self.read(file)
        elif isinstance(initial, Bitmap):
            file = BytesIO(bytes(initial))
            self.read(file)
        elif initial is not None:
            self.read(initial)

    def __repr__(self) -> str:
        tokens = [
            type(self).__name__,
            f"dim={self.dib.width}x{self.dib.height}",
            f"bpp={self.dib.bits_per_pixel}",
            f"comp=0x{self.dib.compression_method:02x}",
            f"pal={self.palette_type()!s}",
        ]
        return "<" + " ".join(tokens) + ">"

    #####################################
    # Getters/Setters for binary fields #
    #####################################

    @property
    def header(self) -> bytes:
        return struct.pack(
            "<2sIHHI",
            self.magic,
            self.filesize,
            self.reserved_a,
            self.reserved_b,
            self.offset,
        )

    @header.setter
    def header(self, value: bytes) -> None:
        (
            self.magic,
            self._filesize,
            self.reserved_a,
            self.reserved_b,
            self._offset,
        ) = struct.unpack("<2sIHHI", value)

    @property
    def filesize(self) -> int:
        if not self._filesize:
            self._filesize = self.offset + len(self.data)
        return self._filesize

    @property
    def offset(self) -> int:
        if not self._offset:
            self._offset = self._HEADER_SIZE + self.dib.size + len(self.palette)
        return self._offset

    @property
    def palette(self) -> bytes:
        return self._palette

    @palette.setter
    def palette(self, value: bytes) -> None:
        if self._invalidate:
            self._offset = 0
            self._filesize = 0
            self._img = None
        self._palette = value

    @property
    def data(self) -> bytes:
        return self._data

    @data.setter
    def data(self, value: bytes) -> None:
        if self._invalidate:
            self._filesize = 0
            self._img = None
        self._data = value

    ####################################################
    # Convenience functions / derived property getters #
    ####################################################

    @property
    def dimensions(self) -> Tuple[int, int]:
        return (self.dib.width, self.dib.height)

    @property
    def row_size(self) -> int:
        return self.dib.row_size()

    def palette_type(self) -> PaletteType:
        if self.palette:
            return PaletteType.EMBEDDED
        return PaletteType.NONE

    def __bytes__(self) -> bytes:
        """
        Return a BMP in its original disk format. (Windows or MSCC compressed)
        """
        return self.bitmap()

    def bitmap(self) -> bytes:
        """
        Return a standard Windows BMP as bytes.
        """
        return self.header + bytes(self.dib) + self.palette + self.data

    def characterize(self, extra_fields: Tuple[Any, ...] = ()) -> Tuple[Any, ...]:
        """
        Debugging convenience method; please don't rely on this as API.

        Return some fields that help characterize what type
        of image this is. Only used for debugging purposes in
        strictness_check() methods in various other classes.

        Returns a Tuple with the following fields:
          - int: color depth (bits per pixel)
          - bool: if image size is present in the DIB header
          - bool: if # colors is non-zero in the DIB header
          - int: the *bmp* (not cchat!) compression method used, if any
          - str: The method that the palette was stored
          - Any extra fields passed in by the caller.
        """
        info = (
            self.dib.bits_per_pixel,
            bool(self.dib.image_size),
            bool(self.dib.n_colors),
            self.dib.compression_method,
            str(self.palette_type()),
        ) + tuple(extra_fields)
        return info

    ##########################
    # Stuff that does stuff! #
    ##########################

    def read(self, file: BinaryIO) -> None:
        self._img = None
        with self._disable_invalidation():
            self._read_header(file)
            self.dib.read(file)
            self._read_palette(file)

            nbytes = self.filesize - self.offset
            data = file.read(nbytes)

            if len(data) != nbytes:
                raise EOFError(f"Expected {nbytes} of bitmap data, but got only {len(data)}")

            self.data = data

    def get_img(self, hack_2bpp_to_1bpp: bool = False) -> Image:
        """
        Get a PIL Image object from this Bitmap.

        :param hack_2bpp_to_1bpp: Due to a bug in some Avatar upgrade
           software, some monochrome images were inadvertantly stored
           as 2bpp. It's impossible to tell from looking at the bitmap
           data alone if this should be corrected, so the caller needs
           to pass this flag in when the correction is desired.
        """
        if self._img:
            return self._img

        # CChat 2.1 files use RLE_4 compression sometimes, which PIL doesn't support. :(
        if self.dib.compression_method == 0x02:
            self._img = self._rle4_decompress().get_img()
            return self._img

        # PIL doesn't support 2-bits-per-pixel. ;_;
        if self.dib.bits_per_pixel == 2:
            self._img = self._2bpp_image(hack_2bpp_to_1bpp)
            return self._img

        # Everything else should be pretty normal.
        self._img = image_open(BytesIO(self.bitmap()))
        return self._img

    ####################
    # Internal helpers #
    ####################

    def _read_header(self, file: BinaryIO) -> None:
        data = file.read(2)
        if data != b"BM":
            raise ValueError(f"Expected Bitmap file header 'BM', got '{data.hex()}'")
        file.seek(-2, os.SEEK_CUR)
        self.header = file.read(self._HEADER_SIZE)

    def _read_palette(self, file: BinaryIO) -> None:
        if self.dib.bits_per_pixel <= 8:
            colors = self.dib.n_colors
            if not colors:
                colors = 2**self.dib.bits_per_pixel
            self.palette = file.read(4 * colors)
        else:
            assert self.dib.n_colors == 0, "Expected dib.n_colors to be 0 when bpp > 8"

    def _2bpp_image(self, hack_2bpp_to_1bpp: bool) -> Image:
        """
        Get a PIL image for 2bpp bitmaps.
        """
        assert self.dib.bits_per_pixel == 2
        if self.dib.compression_method:
            raise NotImplementedError(
                f"DIB compression method {self.dib.compression_method}"
                " not currently supported for 2bpp images."
            )
        newdata = bytearray()
        if hack_2bpp_to_1bpp:
            # This image is 2 bits per pixel, but it was meant to be 1bpp.
            # As a hack, force it back to 1bpp!
            for idx in self._2bpp_indices():
                idx_a = (idx >> 1) * 3
                idx_b = (idx & 0x01) * 3
                newdata.extend(self.palette[4 * idx_a : 4 * idx_a + 4])
                newdata.extend(self.palette[4 * idx_b : 4 * idx_b + 4])
            dimensions = (self.dimensions[0] * 2, self.dimensions[1])
        else:
            # Normal 2bpp image. Well, as normal as those get, anyway. O:-)
            for idx in self._2bpp_indices():
                newdata.extend(self.palette[4 * idx : 4 * idx + 4])
            dimensions = self.dimensions

        img = frombytes(
            "RGBA",
            dimensions,
            bytes(newdata),
            "raw",
            "RGBA",
            0,
            -1,
        )
        if hack_2bpp_to_1bpp:
            img = img.crop((0, 0, *self.dimensions))
        return img

    def _2bpp_indices(self) -> Iterator[int]:
        """
        This method is a generator that yields a sequence of palette indices.

        PIL doesn't natively support 2bpp images. Fair enough, they're
        weird. I think even Windows 95/98 don't support such a
        format. In CChat avatars, they're only used for v2.5 avatars
        to encode multiple 1bpp images together into one
        pseudo-image. In a rare case, there's a single avatar I know
        of (Y-Guy.AVB) that appears to use a 2bpp image for its icon
        as the result of an automatic v2.5 conversion tool bug.
        """
        for y in range(self.dib.height):
            offset = self.row_size * y
            for x in range(self.dib.width):
                shift = 6 - ((x % 4) * 2)  # 6, 4, 2, 0, ...
                pal_idx = (self.data[offset + int(x / 4)] >> shift) & 0x03
                yield pal_idx

    def _rle4_indices(self) -> bytearray:
        """
        This method returns a sequence of palette indices.

        PIL does not appear to natively support RLE4 compressed
        bitmaps. So, this method will yield a sequence of indices with
        no padding.

        This is unfortunately very messy.

        https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-wmf/73b57f24-6d78-4eeb-9c06-8f892d88f1ab
        """
        pixels = bytearray()
        row = bytearray()
        data = BytesIO(self.data)

        def _end_of_line() -> None:
            # I *think* the EOL marker can "imply" pixels. It
            # certainly can for "DELTA" and "EOB" markers.
            n = self.dib.width - len(row)
            if n < 0:
                # I think sometimes data can get saved in the margin,
                # even though it's not "part of the image". In those
                # cases, we need to actually ignore some pixels. Rude.
                for _ in range(abs(n)):
                    row.pop()
                n = 0
            row.extend(b"\x00" * n)
            pixels.extend(row)
            row.clear()

        while True:
            x, y = data.read(2)
            if x == 0:
                if y == 0:
                    # Encoded mode - End of Line
                    _end_of_line()
                elif y == 1:
                    # Encoded mode - End of Bitmap
                    num_rows = int(len(pixels) / self.dib.width)
                    rows_rem = self.dib.height - num_rows
                    for _ in range(rows_rem):
                        _end_of_line()
                    break
                elif y == 2:
                    # Encoded mode - Delta
                    dx, dy = data.read(2)
                    _LOG.debug(f"Delta - {dx},{dy}")
                    x_offset = len(row) + dx
                    for _ in range(dy):
                        _end_of_line()
                    row.extend(b"\x00" * x_offset)
                else:
                    # Absolute mode - y indicates number of indices to follow.
                    # run length is "aligned on a word boundary".
                    rounded = math.ceil(y / 4) * 2
                    abs_data = data.read(rounded)
                    _LOG.debug(f"Absolute mode: {y} pixels")
                    for i in range(y):
                        bval = abs_data[int(i / 2)]
                        row.append((bval >> (0 if (i % 2) else 4)) & 0x0F)
            else:
                # Encoded mode - alternating colors
                _LOG.debug(f"Encoded mode: {x} pixels")
                for i in range(x):
                    row.append((y >> (0 if (i % 2) else 4)) & 0x0F)

        assert len(pixels) == self.dib.width * self.dib.height
        return pixels

    def _rle4_decompress(self) -> "Bitmap":
        """
        Return a whole new bitmap object that removes RLE4 compression.

        This method is *heinously* inefficient, it is notably the
        slowest image deserialization pathway possible in this
        library. I just ran out of neurons to care about improving it,
        and I was happier to have something that was provably correct.

        Apologies for the functionally correct software.
        """
        assert self.dib.compression_method == 0x02, "compression method is not RLE4"
        assert self.dib.bits_per_pixel == 4, "compression method was RLE4 but bpp wasn't 4?"

        def _pad(data: bytes) -> Iterator[int]:
            for x in range(self.dib.height):
                for y in range(self.dib.width):
                    yield data[x * self.dib.width + y]
                for y in range((self.row_size * 2) - self.dib.width):
                    yield 0

        def _pack(it: Iterator[int]) -> Iterator[int]:
            while True:
                try:
                    value = (next(it) << 4) | next(it)
                    yield value
                except StopIteration:
                    break

        tmp_data = bytes(_pack(_pad(self._rle4_indices())))
        assert len(tmp_data) == self.dib.expected_size()

        tmp = Bitmap(self)  # Make a copy!
        tmp.dib.image_size = 0  # This is legal to zero when ...
        tmp.dib.compression_method = 0  # ... compression method is 0.
        tmp.data = tmp_data  # Replace with the uncompressed data O:-)
        return tmp

    @contextmanager
    def _disable_invalidation(self) -> Iterator[None]:
        """
        Turn off cached field invalidation temporarily.

        During initial object construction from existing bitmap data,
        we'll read in filesize and offset. When palette and bitmap
        data are read in, this would otherwise delete the stored
        values. This bypasses that.
        """
        self._invalidate = False
        try:
            yield
        finally:
            self._invalidate = True

        # Paranoia: what if the input data was *bad*?

        given = self._filesize
        self._filesize = 0
        computed = self.filesize

        if given != computed:
            raise ValueError(
                f"Given bitmap filesize {given} does not match computed size {computed}"
            )

        given = self._offset
        self._offset = 0
        computed = self.offset

        if given != computed:
            raise ValueError(f"Given bitmap offset {given} does not match computed size {computed}")


class DIBHeader:
    """
    DIBHeader is a fairly straightforward data class representing the DIB header.

    https://en.wikipedia.org/wiki/BMP_file_format#DIB_header_(bitmap_information_header)

    This class represents only the 40 byte BITMAPINFOHEADER variant.

    Both CChat 2.1 and 2.5 files utilize this structure. See `Bitmap`
    or `CompressedBitmap` for more details.
    """

    _expected_size = 40

    def __init__(self) -> None:
        self.size = self._expected_size
        self.width = 0
        self.height = 0
        self.n_color_planes = 0
        self.bits_per_pixel = 0
        self.compression_method = 0
        self.image_size = 0
        self.x_res = 0
        self.y_res = 0
        self.n_colors = 0
        self.imp_colors = 0

    def read(self, file: BinaryIO) -> None:
        self.size = struct.unpack("<I", file.read(4))[0]
        if self.size != self._expected_size:
            raise ValueError(
                f"Expected {self._expected_size} byte bitmap DIB header, but got size={self.size}"
            )

        data = struct.unpack("<iiHHIIiiII", file.read(36))
        self.width = data[0]
        self.height = data[1]
        self.n_color_planes = data[2]
        self.bits_per_pixel = data[3]
        self.compression_method = data[4]
        self.image_size = data[5]
        self.x_res = data[6]
        self.y_res = data[7]
        self.n_colors = data[8]
        self.imp_colors = data[9]

        # image_size is an optional field when compression is zero,
        # because it can be computed. If it's provided, though, double-check it:
        if self.compression_method == 0 and self.image_size:
            if self.expected_size() != self.image_size:
                _LOG.warning(
                    f"Non-zero DIB field image_size ({self.image_size})"
                    f" does not match expected size ({self.expected_size()}"
                )

    def row_size(self) -> int:
        """Returns the row size of this image in bytes"""
        return math.ceil((self.bits_per_pixel * self.width) / 32) * 4

    def expected_size(self) -> int:
        """Returns the anticipated size of this image in bytes"""
        return self.row_size() * self.height

    def __bytes__(self) -> bytes:
        return struct.pack(
            "<IiiHHIIiiII",
            self.size,
            self.width,
            self.height,
            self.n_color_planes,
            self.bits_per_pixel,
            self.compression_method,
            self.image_size,
            self.x_res,
            self.y_res,
            self.n_colors,
            self.imp_colors,
        )

    def print(self) -> None:
        print(f"  Width: {self.width}")
        print(f"  Height: {self.height}")
        print(f"  Color Planes: {self.n_color_planes}")
        print(f"  Color Depth: {self.bits_per_pixel}")
        print(f"  Compression Method: {self.compression_method}")
        print(f"  Image size: {self.image_size}")
        print(f"  X-Res: {self.x_res}")
        print(f"  Y-Res: {self.y_res}")
        print(f"  # Colors: {self.n_colors}")
        print(f"  # Imp. Colors: {self.imp_colors}")


class CompressedBitmap(Bitmap):
    """
    This class represents the Compressed Bitmaps used in CChat 2.5 files.

    These bitmaps differ from traditional bitmaps by omitting the file
    header completely, loading the palette data *before* the DIB
    header, and using ZLIB to compress the pixel data.

    The structure is:

    - Palette data, if present.
      The palette structure can be present, but empty.
      If the palette data is absent, it means we have an implied palette.
    - DIB header (40 bytes)
    - Uncompressed size (4 bytes)
    - Compressed size (4 bytes)
    - ZLIB compressed pixel data (variable)

    .. note:: Currently, it's not an intentionally supported operation
       to modify the compressed_size, uncompressed_size, cpalette or
       zdata fields directly. Do so at your own risk.

       Setting the .palette or .data fields will update these fields
       accordingly.
    """

    def __init__(
        self,
        initial: Optional[Union[bytes, str, Path, "Bitmap", BinaryIO]] = None,
        autorepair: bool = False,
    ):
        self.compressed_size = 0
        self.uncompressed_size = 0
        self.cpalette: Optional[bytes] = None
        self.zdata: bytes = b""

        # Super call last because it invokes read()
        super().__init__(initial, autorepair)

    #####################################
    # Getters/Setters for binary fields #
    #####################################

    @property
    def palette(self) -> bytes:
        return self._palette

    @palette.setter
    def palette(self, value: bytes) -> None:
        cpalette = bytearray()
        buf = BytesIO(value)
        for _ in range(len(value) // 4):
            cpalette.extend(bytes(reversed(buf.read(3))))
            buf.read(1)
        if self._invalidate:
            self._offset = 0
            self._filesize = 0
            self._img = None
        self.cpalette = bytes(cpalette)
        self._palette = value

    @property
    def data(self) -> bytes:
        if self._data:
            return self._data

        try:
            self._data = zlib.decompress(self.zdata, bufsize=self.uncompressed_size)
        except zlib.error as exc:
            if not self._autorepair:
                raise
            _LOG.info(f"zlib data is corrupted. attempting repair. error: {str(exc)}")
            if not self._repair():
                _LOG.info("zlib data recovery failed.")
                raise
            _LOG.warning("zlib data is corrupted, but was successfully recovered.")

        if len(self._data) != self.uncompressed_size:
            raise ValueError(
                f"Expected {self.uncompressed_size} bytes after decompressing, "
                f"got {len(self._data)} bytes"
            )

        return self._data

    @data.setter
    def data(self, value: bytes) -> None:
        if len(value) != self.dib.expected_size():
            raise ValueError(
                f"Size of new data {len(value)} "
                f"does not match expected size from DIB header {self.dib.expected_size()}"
            )

        if self._invalidate:
            self._filesize = 0
            self._img = None

        zdata = zlib.compress(value)
        self._data = value
        self.zdata = zdata
        self.uncompressed_size = len(value)
        self.compressed_size = len(zdata)

    ####################################################
    # Convenience functions / derived property getters #
    ####################################################

    def __bytes__(self) -> bytes:
        """
        Return the MSCC compressed bitmap bytes for this object.

        Unlike the base class, we want a packing that's in the
        compressed form here.  use bitmap() to get bytes for a normal
        bitmap file instead.

        Though, keep in mind: this direct conversion may not make
        sense for implicit palette bitmaps, which need to be split
        apart and reconstructed into several constituent monochrome
        images instead.
        """
        buff = bytearray()
        if self.cpalette is not None:
            magic = b"\x01\x01"
            num_colors = int(len(self.cpalette) / 3)
            color_length = len(self.cpalette) + 2
            buff += struct.pack("<2sHH", magic, color_length, num_colors)
            buff += self.cpalette

        buff += bytes(self.dib)
        buff += struct.pack("<II", self.uncompressed_size, self.compressed_size)
        buff += self.zdata

        return bytes(buff)

    def palette_type(self) -> PaletteType:
        # Compressed palette is present and non-empty. This occurs for
        # 4bpp and 8bpp base images.
        # imgflag is always 0x02.
        if self.cpalette:
            return PaletteType.EXPLICIT

        # A custom palette header was present, but it didn't contain
        # anything. This appears to occur for 24bps full-color base
        # images that do not use a palette.
        # imgflag is always 0x02.
        if self.cpalette is not None:
            return PaletteType.EMPTY

        # No custom palette was present at all. These images generally
        # contain one or more alpha transparency compositing masks,
        # but can contain monochrome base image data when possible,
        # too. (Primarily for MS-authored comic avatar files.)
        #
        # imgflag == 0x04 (monochrome base img + halo + mask)
        # imgflag == 0x05 (compositing mask for full-color head images)
        # imgflag == 0x03 (halo mask for pre-composited full-color images)
        #
        # See generate_implied_palette() for even more detail.
        return PaletteType.IMPLICIT

    ##########################
    # Stuff that does stuff! #
    ##########################

    def generate_implied_palette(self, flag: int = 0) -> None:
        if self.palette_type() != PaletteType.IMPLICIT:
            return

        assert self.palette == b"", "Unexpected palette data for implicit palette image"

        if self.dib.bits_per_pixel == 1:
            assert flag in (
                0x03,
                0x04,
            ), f"implicit 1bpp flag was 0x{flag:02x}, expected 0x03/0x04"

            if flag == 0x04:
                # 1bpp and flag==0x04 only happens for icons - and I
                # think this is actually just an accident; I think the
                # conversion tool to upgrade avatars to the 2.5 format
                # accidentally compresses monochrome icons in a
                # strange way that was only meant for mask images.

                # I believe flag==0x04 was meant to be for 2bpp
                # images; and in fact when these icons are displayed
                # in cchat they appear at half-width because they are
                # being interpreted as 2bpp images.

                # For our purposes, though, we can just set a special
                # palette for these types of images and the icons will
                # extract correctly based on their original intent
                # (How they appear in v2.1 files), and at the right
                # size.

                # This impacts four avatars I've found: John Bruton,
                # Jean-Luc Dehaene, Costas Simitis and Blues Elmwood.
                self._palette = b"\x00\x00\x00\xFF\xFF\xFF\xFF\xFF"
            elif flag == 0x03:
                # This is used for offset3/halo mask images.
                # It's used in thousands of avatars.
                # The colors match those used in v2.1 mask images.
                self._palette = b"\xFF\xFF\xFF\xFF\x00\x00\x00\xFF"

        elif self.dib.bits_per_pixel == 2:
            assert flag in (
                0x04,
                0x05,
            ), f"implicit 2bpp flag was 0x{flag:02x}, expected 0x04/0x05"
            if flag == 0x05:
                # This is a 2-color mask stored in offset2 for
                # full-color dynamic format avatars.

                # I've found only 11 avatars that use this mode: BUCK,
                # GESSELM, Kalana, McGraw, Ray, Re-Man, Tom, Trisha,
                # Trudy, VERONICA, and zar-zoon.

                # Almost always, they use only the first and last
                # palette entries and behave like normal binary
                # masks. On rare occasion, just a few pixels in one or
                # two poses use the other palette entries; this rarity
                # occurs in BUCK, GESSELM and VERONICA avatars.

                # Behavior for the "rare" palette entries was verified
                # by hand in CChat 2.5 by modifying avatars and
                # observing the behavior.

                self._palette = (
                    b"\xFF\xFF\xFF\xFF"  # White (Transparent)
                    b"\xFF\xFF\xFF\xFF"  # (rarely used - same as above)
                    b"\x00\x00\x00\xFF"  # (rarely used - same as below)
                    b"\x00\x00\x00\xFF"  # Black (Opaque)
                )
            elif flag == 0x04:
                # This is a 4-color image that encodes a monochrome
                # base image, the halo, and the mask. (The halo may be
                # omitted in some circumstances.) The colors I choose
                # below are fully arbitrary, but I am abusing the RGB
                # channels to encode three separate two-color images
                # that we'll extract later with Image.getchannel().

                # Red channel encodes the black and white base image,
                # which interprets the halo and background as fully
                # white.

                # Green channel encodes the transparency mask, which
                # interprets the background and halo region as fully
                # white and the base image pixels as black. This mask
                # will be inverted before PIL uses it as a mask, but
                # it maintains consistency with how masks are stored
                # in 2.1 images.

                # Blue channel encodes the halo mask, which functions
                # similarly to the tight transparency mask, but
                # toggles the halo pixels to black, leaving only the
                # background region white. Like the tight transparency
                # mask, these colors are inverted from PIL's
                # expectation of a transparency mask, but it's
                # consistent with explicit palette data from CChat
                # 2.1.
                self._palette = (
                    b"\xFF\xFF\xFF\xFF"  # Background region
                    b"\xFF\xFF\x00\xFF"  # Halo region
                    b"\xFF\x00\x00\xFF"  # White opaque pixels (base image)
                    b"\x00\x00\x00\xFF"  # Black opaque pixels (base image)
                )

        else:
            raise NotImplementedError(
                f"Cannot generate implicit palette for bpp={self.dib.bits_per_pixel}"
            )

    def read(self, file: BinaryIO) -> None:
        # Compressed bitmaps have no 14 byte header; as a result,
        # filesize and offset are computed only on-demand instead.

        # Weird proprietary palette is *before* the DIB header:
        self._read_custom_palette(file)

        # DIB header is simply the normal kind:
        self.dib.read(file)

        # The palette which would otherwise occur here has already been read.

        (self.uncompressed_size, self.compressed_size) = struct.unpack("<II", file.read(8))

        if self.uncompressed_size != self.dib.expected_size():
            raise ValueError(
                f"Uncompressed size written before ZLIB data ({self.uncompressed_size})"
                f" does not match expected size ({self.dib.expected_size()}), "
                "this bitmap header may be corrupt."
            )
        if self.compressed_size > self.uncompressed_size + 11:
            raise ValueError(
                f"Compressed size written before ZLIB data ({self.compressed_size}) "
                f"is more than 11 bytes larger than the uncompressed/expected size"
                f" ({self.uncompressed_size}), this bitmap header may be corrupt."
            )

        self.zdata = file.read(self.compressed_size)

        if len(self.zdata) != self.compressed_size:
            raise EOFError(
                f"Could not read {self.compressed_size} bytes of compressed data."
                f" (Read only {len(self.zdata)} bytes.)"
            )

    ####################
    # Internal helpers #
    ####################

    def _recover_data_newline_method(self, nbytes: int = 1) -> bool:
        # This method looks at 0x0a bytes and attempts to see if they
        # should have been 0x0a 0x0d sequences. It relies on a human
        # operator having already re-padded the data with null byte(s)
        # to bring it up to "correct size".
        _LOG.debug(f"Attempting to recover data newline repair method (nbytes={nbytes}) ...")
        offset = 0
        indices = []
        while True:
            index = self.zdata.find(b"\x0a", offset)
            if index == -1:
                break
            indices.append(index)
            offset = index + 1

        if not indices:
            _LOG.debug("newline repair method isn't applicable.")
            return False
        _LOG.debug(f"Found {len(indices)} 0x0a bytes")

        attempts = 0
        for configuration in itertools.combinations(indices, nbytes):
            attempts += 1
            zdata = bytearray(self.zdata[:-nbytes])
            for offset in sorted(configuration, reverse=True):
                # Sort backwards so we can insert back-to-front without doing weird math.
                zdata.insert(offset, 0x0D)
            try:
                self._data = zlib.decompress(zdata, bufsize=self.uncompressed_size)
                _LOG.debug(f"\nSuccess!!! - attempt #{attempts}, inserted 0x0D @ {configuration}")
                self.zdata = bytes(zdata)
                return True
            except zlib.error:
                # No good.
                pass

        _LOG.debug(f"\nCouldn't repair using newline repair method (nbytes={nbytes}).")
        return False

    def _repair_tripod(self) -> bool:
        """
        Detect and fix corruption caused by the Tripod webserver.

        The tripod webserver attempted to find and modify apparent
        HTML tags in the response body.

        Generally, this means binary data that happened to look like
        the start of an HTML tag may have been uppercased or
        lowercased, depending.

        In practice, this means data like <b, <f, and <h are
        suspicious in tripod files where the zlib data fails to
        decompress; recovery is sometimes possible by simple
        guess-and-check modifying the capitalization of any such
        bytes.

        Personally, I've not observed any occurrences of anything
        longer than "<BO" or "</B"; three bytes seems about as much as
        you're likely to run into by chance; but who knows! you could
        get very unlucky!
        """
        _LOG.debug("Attempting to repair data using tripod uncorruption method ...")
        tagnames = ("body", "frameset", "html", "head")

        def _make_pattern(tag: str) -> bytes:
            ret = "(?:".join(tag)
            ret += ")?" * (len(tag) - 1)
            return ret.encode()

        patterns = [_make_pattern(tag) for tag in tagnames]
        pattern = b"</?(" + b"|".join(patterns) + b")"

        indices = []
        for match in re.finditer(pattern, self.zdata, re.IGNORECASE):
            _LOG.debug(f"found suspicious data: {match.group(0)!r} @ 0x{match.start(0):02x}")
            for i in range(match.start(1), match.end(1)):
                indices.append(i)

        if not indices:
            _LOG.debug("Tripod data recovery method isn't applicable.")
            return False
        _LOG.debug(f"Found {len(indices)} suspicious bytes")
        _LOG.debug(f"There are {2**len(indices)-1} possible alterations to try")

        zdata = bytearray(self.zdata)

        for configuration in itertools.product((True, False), repeat=len(indices)):
            for offset, modify in zip(indices, configuration):
                if modify:
                    zdata[offset] = ord(chr(zdata[offset]).upper())
                else:
                    zdata[offset] = ord(chr(zdata[offset]).lower())
            try:
                self._data = zlib.decompress(zdata, bufsize=self.uncompressed_size)
                _LOG.debug(f"Success!!! - configuration: {configuration}")
                self.zdata = bytes(zdata)
                return True
            except zlib.error as exc:
                _LOG.debug(f"No good - {str(exc)}")

        _LOG.debug("Couldn't recover the data (tripod method)")
        return False

    def _repair(self) -> bool:
        if self._repair_tripod():
            return True

        # These other methods require knowing how short or overlong
        # the data buffer is, which may not be possible with
        # corruption and may require outside help from the Binary
        # class, so they're stubbed out for now because they aren't
        # fully automatic.

        # if self._recover_data_newline_method(nbytes=1):
        #     return True
        # if self._recover_data_newline_method(nbytes=2):
        #     return True
        return False

    def _read_custom_palette(self, file: BinaryIO) -> None:
        # If there is no palette here, we expect to read '28 00' as
        # the first two bytes of the DIB header. This should be fairly
        # resilient.
        curpos = file.tell()
        magic = file.read(2)
        if magic != b"\x01\x01":
            file.seek(curpos, os.SEEK_SET)
            return

        color_length = struct.unpack("<H", file.read(2))[0]
        if color_length < 2:
            raise ValueError(f"Expected palette size to be two or more bytes, got {color_length}")
        num_colors = struct.unpack("<H", file.read(2))[0]
        if color_length != num_colors * 3 + 2:
            raise ValueError(
                f"Expected palette size to be {num_colors * 3 + 2}, got {color_length}"
            )

        self.cpalette = file.read(num_colors * 3)
        if len(self.cpalette) != (num_colors * 3):
            raise EOFError(
                f"Expected {num_colors * 3} bytes of palette data, read only {len(self.cpalette)}"
            )

        # The packed format here appears to be BGR, Windows BMP files
        # expect RGB0. Reverse the order and pad every three bytes
        # with an empty fourth.
        buf = BytesIO(self.cpalette)
        palette = bytearray()
        for _ in range(num_colors):
            palette.extend(bytes(reversed(buf.read(3))))
            palette.append(0)
        self._palette = bytes(palette)
